"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define ImpactFuncSet class.
"""

__all__ = ['ImpactFuncSet']

import copy
import logging
from itertools import repeat
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import xlsxwriter

from climada.entity.impact_funcs.base import ImpactFunc
from climada.entity.tag import Tag
import climada.util.plot as u_plot
import climada.util.hdf5_handler as u_hdf5

LOGGER = logging.getLogger(__name__)

DEF_VAR_EXCEL = {'sheet_name': 'impact_functions',
                 'col_name': {'func_id': 'impact_fun_id',
                              'inten': 'intensity',
                              'mdd': 'mdd',
                              'paa': 'paa',
                              'name': 'name',
                              'unit': 'intensity_unit',
                              'peril': 'peril_id'
                             }
                }
"""Excel and csv variable names"""

DEF_VAR_MAT = {'sup_field_name': 'entity',
               'field_name': 'damagefunctions',
               'var_name': {'fun_id': 'DamageFunID',
                            'inten': 'Intensity',
                            'mdd': 'MDD',
                            'paa': 'PAA',
                            'name': 'name',
                            'unit': 'Intensity_unit',
                            'peril': 'peril_ID'
                           }
              }
"""MATLAB variable names"""

class ImpactFuncSet():
    """Contains impact functions of type ImpactFunc. Loads from
    files with format defined in FILE_EXT.

    Attributes
    ----------
    tag : Tag
        information about the source data
    _data : dict
        contains ImpactFunc classes. It's not suppossed to be
        directly accessed. Use the class methods instead.
    """

    def __init__(self):
        """Empty initialization.

        Examples
        --------
        Fill impact functions with values and check consistency data:

        >>> fun_1 = ImpactFunc()
        >>> fun_1.haz_type = 'TC'
        >>> fun_1.id = 3
        >>> fun_1.intensity = np.array([0, 20])
        >>> fun_1.paa = np.array([0, 1])
        >>> fun_1.mdd = np.array([0, 0.5])
        >>> imp_fun = ImpactFuncSet()
        >>> imp_fun.append(fun_1)
        >>> imp_fun.check()

        Read impact functions from file and checks consistency data.

        >>> imp_fun = ImpactFuncSet()
        >>> imp_fun.read(ENT_TEMPLATE_XLS)
        """
        self.clear()

    def clear(self):
        """Reinitialize attributes."""
        self.tag = Tag()
        self._data = dict()  # {hazard_type : {id:ImpactFunc}}

    def append(self, func):
        """Append a ImpactFunc. Overwrite existing if same id and haz_type.

        Parameters
        ----------
        func : ImpactFunc
            ImpactFunc instance

        Raises
        ------
        ValueError
        """
        if not isinstance(func, ImpactFunc):
            raise ValueError("Input value is not of type ImpactFunc.")
        if not func.haz_type:
            LOGGER.warning("Input ImpactFunc's hazard type not set.")
        if not func.id:
            LOGGER.warning("Input ImpactFunc's id not set.")
        if func.haz_type not in self._data:
            self._data[func.haz_type] = dict()
        self._data[func.haz_type][func.id] = func

    def remove_func(self, haz_type=None, fun_id=None):
        """Remove impact function(s) with provided hazard type and/or id.
        If no input provided, all impact functions are removed.

        Parameters
        ----------
        haz_type : str, optional
            all impact functions with this hazard
        fun_id : int, optional
            all impact functions with this id
        """
        if (haz_type is not None) and (fun_id is not None):
            try:
                del self._data[haz_type][fun_id]
            except KeyError:
                LOGGER.warning("No ImpactFunc with hazard %s and id %s.",
                               haz_type, fun_id)
        elif haz_type is not None:
            try:
                del self._data[haz_type]
            except KeyError:
                LOGGER.warning("No ImpactFunc with hazard %s.", haz_type)
        elif fun_id is not None:
            haz_remove = self.get_hazard_types(fun_id)
            if not haz_remove:
                LOGGER.warning("No ImpactFunc with id %s.", fun_id)
            for vul_haz in haz_remove:
                del self._data[vul_haz][fun_id]
        else:
            self._data = dict()

    def get_func(self, haz_type=None, fun_id=None):
        """Get ImpactFunc(s) of input hazard type and/or id.
        If no input provided, all impact functions are returned.

        Parameters
        ----------
        haz_type : str, optional
            hazard type
        fun_id : int, optional
            ImpactFunc id

        Returns
        -------
        ImpactFunc (if haz_type and fun_id),
        list(ImpactFunc) (if haz_type or fun_id),
        {ImpactFunc.haz_type : {ImpactFunc.id : ImpactFunc}} (if None)
        """
        if (haz_type is not None) and (fun_id is not None):
            try:
                return self._data[haz_type][fun_id]
            except KeyError:
                return list()
        elif haz_type is not None:
            try:
                return list(self._data[haz_type].values())
            except KeyError:
                return list()
        elif fun_id is not None:
            haz_return = self.get_hazard_types(fun_id)
            vul_return = []
            for vul_haz in haz_return:
                vul_return.append(self._data[vul_haz][fun_id])
            return vul_return
        else:
            return self._data

    def get_hazard_types(self, fun_id=None):
        """Get impact functions hazard types contained for the id provided.
        Return all hazard types if no input id.

        Parameters
        ----------
        fun_id : int, optional
            id of an impact function

        Returns
        -------
        list(str)
        """
        if fun_id is None:
            return list(self._data.keys())

        haz_types = []
        for vul_haz, vul_dict in self._data.items():
            if fun_id in vul_dict:
                haz_types.append(vul_haz)
        return haz_types

    def get_ids(self, haz_type=None):
        """Get impact functions ids contained for the hazard type provided.
        Return all ids for each hazard type if no input hazard type.

        Parameters
        ----------
        haz_type : str, optional
            hazard type from which to obtain the ids

        Returns
        -------
        list(ImpactFunc.id) (if haz_type provided),
        {ImpactFunc.haz_type : list(ImpactFunc.id)} (if no haz_type)
        """
        if haz_type is None:
            out_dict = dict()
            for vul_haz, vul_dict in self._data.items():
                out_dict[vul_haz] = list(vul_dict.keys())
            return out_dict

        try:
            return list(self._data[haz_type].keys())
        except KeyError:
            return list()

    def size(self, haz_type=None, fun_id=None):
        """Get number of impact functions contained with input hazard type and
        /or id. If no input provided, get total number of impact functions.

        Parameters
        ----------
        haz_type : str, optional
            hazard type
        fun_id : int, optional
            ImpactFunc id

        Returns
        -------
        int
        """
        if (haz_type is not None) and (fun_id is not None) and \
        (isinstance(self.get_func(haz_type, fun_id), ImpactFunc)):
            return 1
        if (haz_type is not None) or (fun_id is not None):
            return len(self.get_func(haz_type, fun_id))
        return sum(len(vul_list) for vul_list in self.get_ids().values())

    def check(self):
        """Check instance attributes.

        Raises
        ------
        ValueError
        """
        for key_haz, vul_dict in self._data.items():
            for fun_id, vul in vul_dict.items():
                if (fun_id != vul.id) | (fun_id == ''):
                    raise ValueError("Wrong ImpactFunc.id: %s != %s."
                                     % (fun_id, vul.id))
                if (key_haz != vul.haz_type) | (key_haz == ''):
                    raise ValueError("Wrong ImpactFunc.haz_type: %s != %s."
                                     % (key_haz, vul.haz_type))
                vul.check()

    def extend(self, impact_funcs):
        """Append impact functions of input ImpactFuncSet to current
        ImpactFuncSet. Overwrite ImpactFunc if same id and haz_type.

        Parameters
        ----------
        impact_funcs : ImpactFuncSet
            ImpactFuncSet instance to extend

        Raises
        ------
        ValueError
        """
        impact_funcs.check()
        if self.size() == 0:
            self.__dict__ = copy.deepcopy(impact_funcs.__dict__)
            return

        self.tag.append(impact_funcs.tag)

        new_func = impact_funcs.get_func()
        for _, vul_dict in new_func.items():
            for _, vul in vul_dict.items():
                self.append(vul)

    def plot(self, haz_type=None, fun_id=None, axis=None, **kwargs):
        """Plot impact functions of selected hazard (all if not provided) and
        selected function id (all if not provided).

        Parameters
        ----------
        haz_type : str, optional
            hazard type
        fun_id : int, optional
            id of the function

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        num_plts = self.size(haz_type, fun_id)
        num_row, num_col = u_plot._get_row_col_size(num_plts)
        # Select all hazard types to plot
        if haz_type is not None:
            hazards = [haz_type]
        else:
            hazards = self._data.keys()

        if not axis:
            _, axis = plt.subplots(num_row, num_col)
        if num_plts > 1:
            axes = axis.flatten()
        else:
            axes = [axis]

        i_axis = 0
        for sel_haz in hazards:
            if fun_id is not None:
                self._data[sel_haz][fun_id].plot(axis=axes[i_axis], **kwargs)
                i_axis += 1
            else:
                for sel_id in self._data[sel_haz].keys():
                    self._data[sel_haz][sel_id].plot(axis=axes[i_axis], **kwargs)
                    i_axis += 1
        return axis

    @classmethod
    def from_excel(cls, file_name, description='', var_names=DEF_VAR_EXCEL):
        """Read excel file following template and store variables.

        Parameters
        ----------
        file_name : str
            absolute file name
        description : str, optional
            description of the data
        var_names : dict, optional
            name of the variables in the file

        Returns
        -------
        ImpFuncSet
        """
        imp_func_set = cls()
        dfr = pd.read_excel(file_name, var_names['sheet_name'])

        imp_func_set.clear()
        imp_func_set.tag.file_name = str(file_name)
        imp_func_set.tag.description = description
        imp_func_set._fill_dfr(dfr, var_names)
        return imp_func_set

    def read_excel(self, *args, **kwargs):
        """This function is deprecated, use ImpactFuncSet.from_excel instead."""
        LOGGER.warning("The use of ImpactFuncSet.read_excel is deprecated." +
                        "Use ImpactFuncSet.from_excel instead.")
        self.__dict__ = ImpactFuncSet.from_excel(*args, **kwargs).__dict__

    @classmethod
    def from_mat(cls, file_name, description='', var_names=DEF_VAR_MAT):
        """Read MATLAB file generated with previous MATLAB CLIMADA version.

        Parameters
        ----------
        file_name : str
            absolute file name
        description : str, optional
            description of the data
        var_names : dict, optional
            name of the variables in the file

        Return
        ------
        impf_set : climada.entity.impact_func_set.ImpactFuncSet
            Impact func set as defined in matlab file.
        """
        def _get_hdf5_funcs(imp, file_name, var_names):
            """Get rows that fill every impact function and its name."""
            func_pos = dict()
            for row, (fun_id, fun_type) in enumerate(
                    zip(imp[var_names['var_name']['fun_id']].squeeze(),
                        imp[var_names['var_name']['peril']].squeeze())):
                type_str = u_hdf5.get_str_from_ref(file_name, fun_type)
                key = (type_str, int(fun_id))
                if key not in func_pos:
                    func_pos[key] = list()
                func_pos[key].append(row)
            return func_pos

        def _get_hdf5_str(imp, idxs, file_name, var_name):
            """Get rows with same string in var_name."""
            prev_str = ""
            for row in idxs:
                cur_str = u_hdf5.get_str_from_ref(file_name, imp[var_name][row][0])
                if prev_str == "":
                    prev_str = cur_str
                elif prev_str != cur_str:
                    raise ValueError("Impact function with two different %s." % var_name)
            return prev_str

        imp = u_hdf5.read(file_name)
        impf_set = cls()
        impf_set.tag.file_name = str(file_name)
        impf_set.tag.description = description

        try:
            imp = imp[var_names['sup_field_name']]
        except KeyError:
            pass
        try:
            imp = imp[var_names['field_name']]
            funcs_idx = _get_hdf5_funcs(imp, file_name, var_names)
            for imp_key, imp_rows in funcs_idx.items():
                func = ImpactFunc()
                func.haz_type = imp_key[0]
                func.id = imp_key[1]
                # check that this function only has one intensity unit, if provided
                try:
                    func.intensity_unit = _get_hdf5_str(imp, imp_rows,
                                                        file_name,
                                                        var_names['var_name']['unit'])
                except KeyError:
                    pass
                # check that this function only has one name
                try:
                    func.name = _get_hdf5_str(imp, imp_rows, file_name,
                                              var_names['var_name']['name'])
                except KeyError:
                    func.name = str(func.id)
                func.intensity = np.take(imp[var_names['var_name']['inten']], imp_rows)
                func.mdd = np.take(imp[var_names['var_name']['mdd']], imp_rows)
                func.paa = np.take(imp[var_names['var_name']['paa']], imp_rows)
                impf_set.append(func)
        except KeyError as err:
            raise KeyError("Not existing variable: %s" % str(err)) from err

        return impf_set

    def read_mat(self, *args, **kwargs):
        """This function is deprecated, use ImpactFuncSet.from_mat instead."""
        LOGGER.warning("The use of ImpactFuncSet.read_mat  is deprecated."
                       "Use ImpactFuncSet.from_mat  instead.")
        self.__dict__ = ImpactFuncSet.from_mat(*args, **kwargs).__dict__

    def write_excel(self, file_name, var_names=DEF_VAR_EXCEL):
        """Write excel file following template.

        Parameters
        ----------
        file_name : str
            absolute file name to write
        var_names : dict, optional
            name of the variables in the file
        """
        def write_impf(row_ini, imp_ws, xls_data):
            """Write one impact function"""
            for icol, col_dat in enumerate(xls_data):
                for irow, data in enumerate(col_dat, row_ini):
                    imp_ws.write(irow, icol, data)

        imp_wb = xlsxwriter.Workbook(file_name)
        imp_ws = imp_wb.add_worksheet(var_names['sheet_name'])

        header = [var_names['col_name']['func_id'], var_names['col_name']['inten'],
                  var_names['col_name']['mdd'], var_names['col_name']['paa'],
                  var_names['col_name']['peril'], var_names['col_name']['unit'],
                  var_names['col_name']['name']]
        for icol, head_dat in enumerate(header):
            imp_ws.write(0, icol, head_dat)
        row_ini = 1
        for fun_haz_id, fun_haz in self._data.items():
            for fun_id, fun in fun_haz.items():
                n_inten = fun.intensity.size
                xls_data = [repeat(fun_id, n_inten), fun.intensity, fun.mdd,
                            fun.paa, repeat(fun_haz_id, n_inten),
                            repeat(fun.intensity_unit, n_inten),
                            repeat(fun.name, n_inten)]
                write_impf(row_ini, imp_ws, xls_data)
                row_ini += n_inten
        imp_wb.close()

    def _fill_dfr(self, dfr, var_names):

        def _get_xls_funcs(dfr, var_names):
            """Parse individual impact functions."""
            dist_func = []
            for (haz_type, imp_id) in zip(dfr[var_names['col_name']['peril']],
                                          dfr[var_names['col_name']['func_id']]):
                if (haz_type, imp_id) not in dist_func:
                    dist_func.append((haz_type, imp_id))
            return dist_func

        try:
            dist_func = _get_xls_funcs(dfr, var_names)
            for haz_type, imp_id in dist_func:
                df_func = dfr[dfr[var_names['col_name']['peril']] == haz_type]
                df_func = df_func[df_func[var_names['col_name']['func_id']]
                                  == imp_id]

                func = ImpactFunc()
                func.haz_type = haz_type
                func.id = imp_id
                # check that the unit of the intensity is the same
                try:
                    if len(df_func[var_names['col_name']['name']].unique()) != 1:
                        raise ValueError('Impact function with two different names.')
                    func.name = df_func[var_names['col_name']['name']].values[0]
                except KeyError:
                    func.name = str(func.id)

                # check that the unit of the intensity is the same, if provided
                try:
                    if len(df_func[var_names['col_name']['unit']].unique()) != 1:
                        raise ValueError('Impact function with two different \
                                         intensity units.')
                    func.intensity_unit = \
                                    df_func[var_names['col_name']['unit']].values[0]
                except KeyError:
                    pass

                func.intensity = df_func[var_names['col_name']['inten']].values
                func.mdd = df_func[var_names['col_name']['mdd']].values
                func.paa = df_func[var_names['col_name']['paa']].values

                self.append(func)

        except KeyError as err:
            raise KeyError("Not existing variable: %s" % str(err)) from err



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define InputVar class.
"""

import copy
from functools import partial
from itertools import zip_longest
import logging

import scipy as sp
import numpy as np
import matplotlib.pyplot as plt

from climada.entity import Entity, DiscRates

LOGGER = logging.getLogger(__name__)

__all__ = ['InputVar']

FIG_W, FIG_H = 8, 5 #default figize width/heigh column/work multiplicators

class InputVar():
    """
    Input variable for the uncertainty analysis

    An uncertainty input variable requires a single or multi-parameter function.
    The parameters must follow a given distribution. The uncertainty input
    variables are the input parameters of the model.

    Attributes
    ----------
    distr_dict : dict
        Distribution of the uncertainty parameters. Keys are uncertainty
        parameters names and Values are probability density distribution
        from the scipy.stats package
        https://docs.scipy.org/doc/scipy/reference/stats.html
    labels : list
        Names of the uncertainty parameters (keys of distr_dict)
    func : function
        User defined python fucntion with the uncertainty parameters
        as keyword arguements and which returns a climada object.

    Notes
    -----
    A few default Variables are defined for Hazards, Exposures,
    Impact Fucntions, Measures and Entities.


    Examples
    --------

    Categorical variable function: LitPop exposures with m,n exponents in [0,5]
        import scipy as sp
        def litpop_cat(m, n):
            exp = Litpop.from_countries('CHE', exponent=[m, n])
            return exp
        distr_dict = {
            'm': sp.stats.randint(low=0, high=5),
            'n': sp.stats.randint(low=0, high=5)
            }
        iv_cat = InputVar(func=litpop_cat, distr_dict=distr_dict)

    Continuous variable function: Impact function for TC
        import scipy as sp
        def imp_fun_tc(G, v_half, vmin, k, _id=1):
            imp_fun = ImpactFunc()
            imp_fun.haz_type = 'TC'
            imp_fun.id = _id
            imp_fun.intensity_unit = 'm/s'
            imp_fun.intensity = np.linspace(0, 150, num=100)
            imp_fun.mdd = np.repeat(1, len(imp_fun.intensity))
            imp_fun.paa = np.array([sigmoid_function(v, G, v_half, vmin, k)
                                    for v in imp_fun.intensity])
            imp_fun.check()
            impf_set = ImpactFuncSet()
            impf_set.append(imp_fun)
            return impf_set
        distr_dict = {"G": sp.stats.uniform(0.8, 1),
              "v_half": sp.stats.uniform(50, 100),
              "vmin": sp.stats.norm(loc=15, scale=30),
              "k": sp.stats.randint(low=1, high=9)
              }
        iv_cont = InputVar(func=imp_fun_tc, distr_dict=distr_dict)

    """

    def __init__(self, func, distr_dict):
        """
        Initialize InputVar

        Parameters
        ----------
        func : function
            Variable defined as a function of the uncertainty parameters
        distr_dict : dict
            Dictionary of the probability density distributions of the
            uncertainty parameters, with keys matching the keyword
            arguments (i.e. uncertainty parameters) of the func
            function.
            The distribution must be of type scipy.stats
            https://docs.scipy.org/doc/scipy/reference/stats.html

        """
        self.labels = list(distr_dict.keys())
        self.distr_dict = distr_dict
        self.func = func

    def evaluate(self, **params):
        """
        Return the value of uncertainty input variable.

        By default, the value of the average is returned.

        Parameters
        ----------
        **params : optional
            Input parameters will be passed to self.InputVar_func.

        Returns
        -------
        unc_func(**params) : climada object
            Output of the uncertainty variable.

        """

        if not params:
            params = {
                param: distr.mean()
                for param, distr in self.distr_dict.items()
                }
        return self.func(**params)


    def plot(self, figsize=None):
        """
        Plot the distributions of the parameters of the uncertainty variable.

        Parameters
        ----------
        figsize: tuple(int or float, int or float), optional
            The figsize argument of matplotlib.pyplot.subplots()
            The default is derived from the total number of plots (nplots) as:
                nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
                figsize = (ncols * FIG_W, nrows * FIG_H)

        Returns
        -------
        axes: matplotlib.pyplot.figure, matplotlib.pyplot.axes
            The figure and axes handle of the plot.

        """

        nplots = len(self.distr_dict)
        nrows, ncols = int(np.ceil(nplots / 3)), min(nplots, 3)
        if figsize is None:
            figsize = (ncols * FIG_W, nrows * FIG_H)
        _fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        if nplots > 1:
            flat_axes = axes.flatten()
        else:
            flat_axes = np.array([axes])
        for ax, name_distr in zip_longest(flat_axes,
                                    self.distr_dict.items(),
                                    fillvalue=None):
            if name_distr is None:
                ax.remove()
                continue
            (param_name, distr) = name_distr
            low = distr.ppf(1e-10)
            high = distr.ppf(1-1e-10)
            n = 100
            try:
                x = np.linspace(low, high, n)
                ax.plot(x, distr.pdf(x), label=param_name)
            except AttributeError:
                if (high - low) > n:
                    x = np.arange(low, high, int((high-low) / n))
                else:
                    x = np.arange(low, high+1)
                ax.vlines(x, 0, distr.pmf(x), label=param_name)
            ax.legend()
        return axes


    @staticmethod
    def var_to_inputvar(var):
        """
        Returns an uncertainty variable with no distribution if var is not
        an InputVar. Else, returns var.

        Parameters
        ----------
        var : climada.uncertainty.InputVar or any other CLIMADA object

        Returns
        -------
        InputVar
            var if var is InputVar, else InputVar with var and no distribution.

        """

        if isinstance(var, InputVar):
            return var

        return InputVar(func=lambda: var, distr_dict={})

    @staticmethod
    def haz(haz, n_ev=None, bounds_int=None, bounds_freq=None):
        """
        Helper wrapper for basic hazard uncertainty input variable

        The following types of uncertainties can be added:
        HE: sub-sampling events from the total event set
            For each sub-sample, n_ev events are sampled with replacement.
            HE is the value of the seed
            for  the uniform random number generator.
        HI: scale the intensity of all events (homogeneously)
            The instensity of all events is multiplied by a number
            sampled uniformly from a distribution with (min, max) = bounds_int
        HF: scale the frequency of all events (homogeneously)
            The frequency of all events is multiplied by a number
            sampled uniformly from a distribution with (min, max) = bounds_freq

        If a bounds is None, this parameter is assumed to have no uncertainty.

        Parameters
        ----------
        haz : climada.hazard.Hazard
            The base hazard
        n_ev : int, optional
            Number of events to be subsampled per sample. Can be equal or
            larger than haz.size. The default is None.
        bounds_int : (float, float), optional
            Bounds of the uniform distribution for the homogeneous intensity
            scaling. The default is None.
        bounds_freq : (float, float), optional
            Bounds of the uniform distribution for the homogeneous frequency
            scaling. The default is None.

        Returns
        -------
        climada.engine.unsequa.input_var.InputVar
            Uncertainty input variable for a hazard object.

        """
        kwargs = {'haz': haz, 'n_ev': n_ev}
        if n_ev is None:
            kwargs['HE'] = None
        if bounds_int is None:
            kwargs['HI'] = None
        if bounds_freq is None:
            kwargs['HF'] = None
        return InputVar(
            partial(_haz_uncfunc, **kwargs),
            _haz_unc_dict(n_ev, bounds_int, bounds_freq)
            )

    @staticmethod
    def exp(exp_list, bounds_totval=None, bounds_noise=None):
        """
        Helper wrapper for basic exposure uncertainty input variable

        The following types of uncertainties can be added:
        ET: scale the total value (homogeneously)
            The value at each exposure point is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_totvalue
        EN: mutliplicative noise (inhomogeneous)
            The value of each exposure point is independently multiplied by
            a random number sampled uniformly from a distribution
            with (min, max) = bounds_noise. EN is the value of the seed
            for  the uniform random number generator.
        EL: sample uniformly from exposure list
            From the provided list of exposure is elements are uniformly
            sampled. For example, LitPop instances with different exponents.

        If a bounds is None, this parameter is assumed to have no uncertainty.

        Parameters
        ----------
        exp_list : list of climada.entity.exposures.Exposures
            The list of base exposure. Can be one or many to uniformly sample
            from.
        bounds_totval : (float, float), optional
            Bounds of the uniform distribution for the homogeneous total value
            scaling.. The default is None.
        bounds_noise : (float, float), optional
            Bounds of the uniform distribution to scale each exposure point
            independently. The default is None.

        Returns
        -------
        climada.engine.unsequa.input_var.InputVar
            Uncertainty input variable for an exposure object.

        """
        n_exp = len(exp_list)
        kwargs = {'exp_list': exp_list, 'bounds_noise': bounds_noise}
        if bounds_noise is None:
            kwargs['EN'] = None
        if bounds_totval is None:
            kwargs['ET'] = None
        if n_exp == 1:
            kwargs['EL'] = 0
        return InputVar(
            partial(_exp_uncfunc, **kwargs),
            _exp_unc_dict(bounds_totval=bounds_totval,
                          bounds_noise=bounds_noise,
                          n_exp=n_exp)
            )

    @staticmethod
    def impfset(impf_set, haz_id_dict= None, bounds_mdd=None, bounds_paa=None,
                    bounds_impfi=None):
        """
        Helper wrapper for basic impact function set uncertainty input variable.

        One impact function (chosen with haz_type and fun_id) is characterized.

        The following types of uncertainties can be added:
        MDD: scale the mdd (homogeneously)
            The value of mdd at each intensity is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_mdd
        PAA: scale the paa (homogeneously)
            The value of paa at each intensity is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_paa
        IFi: shift the intensity (homogeneously)
            The value intensity are all summed with a random number
            sampled uniformly from a distribution with
            (min, max) = bounds_int

        If a bounds is None, this parameter is assumed to have no uncertainty.

        Parameters
        ----------
        impf_set : climada.entity.impact_funcs.impact_func_set.ImpactFuncSet
            The base impact function set.
        bounds_mdd : (float, float), optional
            Bounds of the uniform distribution for the homogeneous mdd
            scaling. The default is None.
        bounds_paa : (float, float), optional
            Bounds of the uniform distribution for the homogeneous paa
            scaling. The default is None.
        bounds_impfi : (float, float), optional
            Bounds of the uniform distribution for the homogeneous shift
            of intensity. The default is None.
        haz_id_dict : dict(), optional
            Dictionary of the impact functions affected by uncertainty.
            Keys are hazard types (str), values are a list of impact
            function id (int).
            Default is impsf_set.get_ids() i.e. all impact functions in the set

        Returns
        -------
        climada.engine.unsequa.input_var.InputVar
            Uncertainty input variable for an impact function set object.

        """
        kwargs = {}
        if bounds_mdd is None:
            kwargs['MDD'] = None
        if bounds_paa is None:
            kwargs['PAA'] = None
        if bounds_impfi is None:
            kwargs['IFi'] = None
        if haz_id_dict is None:
            haz_id_dict = impf_set.get_ids()
        return InputVar(
            partial(
                _impfset_uncfunc, impf_set=impf_set, haz_id_dict=haz_id_dict,
                **kwargs
                ),
            _impfset_unc_dict(bounds_impfi, bounds_mdd, bounds_paa)
        )

    @staticmethod
    def ent(impf_set, disc_rate, exp_list, meas_set, haz_id_dict,
            bounds_disc=None, bounds_cost=None, bounds_totval=None,
            bounds_noise=None, bounds_mdd=None, bounds_paa=None,
            bounds_impfi=None):
        """
        Helper wrapper for basic entity set uncertainty input variable.

        Important: only the impact function defined by haz_type and
        fun_id will be affected by bounds_impfi, bounds_mdd, bounds_paa.

        The following types of uncertainties can be added:
        DR: value of constant discount rate (homogeneously)
            The value of the discounts in each year is
            sampled uniformly from a distribution with
            (min, max) = bounds_disc
        CO: scale the cost (homogeneously)
            The cost of all measures is multiplied by the same number
            sampled uniformly from a distribution with
            (min, max) = bounds_cost
        ET: scale the total value (homogeneously)
            The value at each exposure point is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_totval
        EN: mutliplicative noise (inhomogeneous)
            The value of each exposure point is independently multiplied by
            a random number sampled uniformly from a distribution
            with (min, max) = bounds_noise. EN is the value of the seed
            for  the uniform random number generator.
        EL: sample uniformly from exposure list
            From the provided list of exposure is elements are uniformly
            sampled. For example, LitPop instances with different exponents.
        MDD: scale the mdd (homogeneously)
            The value of mdd at each intensity is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_mdd
        PAA: scale the paa (homogeneously)
            The value of paa at each intensity is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_paa
        IFi: shift the intensity (homogeneously)
            The value intensity are all summed with a random number
            sampled uniformly from a distribution with
            (min, max) = bounds_int


        If a bounds is None, this parameter is assumed to have no uncertainty.


        Parameters
        ----------
        bounds_disk : (float, float), optional
            Bounds of the uniform distribution for the homogeneous discount
            rate scaling. The default is None.
        bounds_cost :(float, float), optional
            Bounds of the uniform distribution for the homogeneous cost
            of all measures scaling. The default is None.
        bounds_totval : (float, float), optional
            Bounds of the uniform distribution for the homogeneous total
            exposure value scaling. The default is None.
        bounds_noise : (float, float), optional
            Bounds of the uniform distribution to scale each exposure point
            independently. The default is None.
        bounds_mdd : (float, float), optional
            Bounds of the uniform distribution for the homogeneous mdd
            scaling. The default is None.
        bounds_paa : (float, float), optional
            Bounds of the uniform distribution for the homogeneous paa
            scaling. The default is None.
        bounds_impfi : (float, float), optional
            Bounds of the uniform distribution for the homogeneous shift
            of intensity. The default is None.
        impf_set : climada.engine.impact_funcs.impact_func_set.ImpactFuncSet
            The base impact function set.
        disc_rate : climada.entity.disc_rates.base.DiscRates
            The base discount rates.
        exp_list : [climada.entity.exposures.base.Exposure]
            The list of base exposure. Can be one or many to uniformly sample
            from.
        meas_set : climada.entity.measures.measure_set.MeasureSet
            The base measures.
        haz_id_dict : dict
            Dictionary of the impact functions affected by uncertainty.
            Keys are hazard types (str), values are a list of impact
            function id (int).

        Returns
        -------
        climada.engine.unsequa.input_var.InputVar
            Entity uncertainty input variable

        """
        n_exp = len(exp_list)

        kwargs = {}
        if bounds_mdd is None:
            kwargs['MDD'] = None
        if bounds_paa is None:
            kwargs['PAA'] = None
        if bounds_impfi is None:
            kwargs['IFi'] = None
        if bounds_disc is None:
            kwargs['DR'] = None
        if bounds_cost is None:
            kwargs['CO'] = None
        if bounds_totval is None:
            kwargs['ET'] = None
        if bounds_noise is None:
            kwargs['EN'] = None
        if n_exp == 1:
            kwargs['EL'] = 0

        return InputVar(
            partial(_ent_unc_func,
                    impf_set=impf_set, haz_id_dict=haz_id_dict,
                    disc_rate=disc_rate, bounds_noise=bounds_noise,
                    exp_list=exp_list, meas_set=meas_set, **kwargs
                    ),
            _ent_unc_dict(bounds_totval=bounds_totval, bounds_noise=bounds_noise,
                          bounds_impfi=bounds_impfi, bounds_mdd=bounds_mdd,
                          bounds_paa=bounds_paa, bounds_disc=bounds_disc,
                          bounds_cost=bounds_cost, n_exp=n_exp)
        )

    @staticmethod
    def entfut(impf_set, exp_list, meas_set, haz_id_dict,
               bounds_cost=None, bounds_eg=None, bounds_noise=None,
               bounds_impfi=None, bounds_mdd=None, bounds_paa=None,
               ):
        """
        Helper wrapper for basic future entity set uncertainty input variable.

        Important: only the impact function defined by haz_type and
        fun_id will be affected by bounds_impfi, bounds_mdd, bounds_paa.

        The following types of uncertainties can be added:
        CO: scale the cost (homogeneously)
            The cost of all measures is multiplied by the same number
            sampled uniformly from a distribution with
            (min, max) = bounds_cost
        EG: scale the exposures growth (homogeneously)
            The value at each exposure point is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_eg
        EN: mutliplicative noise (inhomogeneous)
            The value of each exposure point is independently multiplied by
            a random number sampled uniformly from a distribution
            with (min, max) = bounds_noise. EN is the value of the seed
            for  the uniform random number generator.
        EL: sample uniformly from exposure list
            From the provided list of exposure is elements are uniformly
            sampled. For example, LitPop instances with different exponents.
        MDD: scale the mdd (homogeneously)
            The value of mdd at each intensity is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_mdd
        PAA: scale the paa (homogeneously)
            The value of paa at each intensity is multiplied by a number
            sampled uniformly from a distribution with
            (min, max) = bounds_paa
        IFi: shift the impact function intensity (homogeneously)
            The value intensity are all summed with a random number
            sampled uniformly from a distribution with
            (min, max) = bounds_impfi


        If a bounds is None, this parameter is assumed to have no uncertainty.


        Parameters
        ----------
        bounds_cost :(float, float), optional
            Bounds of the uniform distribution for the homogeneous cost
            of all measures scaling. The default is None.
        bounds_eg : (float, float), optional
            Bounds of the uniform distribution for the homogeneous total
            exposure growth scaling. The default is None.
        bounds_noise : (float, float), optional
            Bounds of the uniform distribution to scale each exposure point
            independently. The default is None.
        bounds_mdd : (float, float), optional
            Bounds of the uniform distribution for the homogeneous mdd
            scaling. The default is None.
        bounds_paa : (float, float), optional
            Bounds of the uniform distribution for the homogeneous paa
            scaling. The default is None.
        bounds_impfi : (float, float), optional
            Bounds of the uniform distribution for the homogeneous shift
            of intensity. The default is None.
        impf_set : climada.engine.impact_funcs.impact_func_set.ImpactFuncSet
            The base impact function set.
        exp_list : [climada.entity.exposures.base.Exposure]
            The list of base exposure. Can be one or many to uniformly sample
            from.
        meas_set : climada.entity.measures.measure_set.MeasureSet
            The base measures.
        haz_id_dict : dict
            Dictionary of the impact functions affected by uncertainty.
            Keys are hazard types (str), values are a list of impact
            function id (int).

        Returns
        -------
        climada.engine.unsequa.input_var.InputVar
            Entity uncertainty input variable

        """
        n_exp = len(exp_list)

        kwargs = {}
        if bounds_mdd is None:
            kwargs['MDD'] = None
        if bounds_paa is None:
            kwargs['PAA'] = None
        if bounds_impfi is None:
            kwargs['IFi'] = None
        if bounds_cost is None:
            kwargs['CO'] = None
        if bounds_eg is None:
            kwargs['EG'] = None
        if bounds_noise is None:
            kwargs['EN'] = None
        if n_exp == 1:
            kwargs['EL'] = 0

        return InputVar(
            partial(_entfut_unc_func, haz_id_dict=haz_id_dict,
                    bounds_noise=bounds_noise, impf_set=impf_set,
                    exp_list=exp_list, meas_set=meas_set, **kwargs),
            _entfut_unc_dict(bounds_eg=bounds_eg, bounds_noise=bounds_noise,
                             bounds_impfi=bounds_impfi, bounds_paa=bounds_paa,
                             bounds_mdd=bounds_mdd, bounds_cost=bounds_cost,
                             n_exp=n_exp)
        )


#Hazard
def _haz_uncfunc(HE, HI, HF, haz, n_ev):
    haz_tmp = copy.deepcopy(haz)
    if HE is not None:
        rng = np.random.RandomState(int(HE))
        event_names = list(rng.choice(haz_tmp.event_name, int(n_ev)))
        haz_tmp = haz_tmp.select(event_names=event_names)
    if HI is not None:
        haz_tmp.intensity = haz_tmp.intensity.multiply(HI)
    if HF is not None:
        haz_tmp.frequency = np.multiply(haz_tmp.frequency, HF)
    return haz_tmp

def _haz_unc_dict(n_ev, bounds_int, bounds_freq):
    hud = {}
    if n_ev is not None:
        hud['HE'] = sp.stats.randint(0, 2**32 - 1) #seed for rnd generator
    if bounds_int is not None:
        imin, idelta = bounds_int[0], bounds_int[1] - bounds_int[0]
        hud['HI'] = sp.stats.uniform(imin, idelta)
    if bounds_freq is not None:
        fmin, fdelta = bounds_freq[0], bounds_freq[1] - bounds_freq[0]
        hud['HF'] = sp.stats.uniform(fmin, fdelta)
    return hud

#Exposure
def _exp_uncfunc(EN, ET, EL, exp_list, bounds_noise):
    exp_tmp = exp_list[int(EL)].copy(deep=True)
    if EN is not None:
        rng = np.random.RandomState(int(EN))
        rnd_vals = rng.uniform(bounds_noise[0], bounds_noise[1], size = len(exp_tmp.gdf))
        exp_tmp.gdf.value *= rnd_vals
    if ET is not None:
        exp_tmp.gdf.value *= ET
    return exp_tmp

def _exp_unc_dict(bounds_totval, bounds_noise, n_exp):
    eud = {}
    if bounds_totval is not None:
        tmin, tmax = bounds_totval[0], bounds_totval[1] - bounds_totval[0]
        eud['ET'] = sp.stats.uniform(tmin, tmax)
    if bounds_noise is not None:
        eud['EN'] = sp.stats.randint(0, 2**32 - 1) #seed for rnd generator
    if n_exp > 1:
        eud['EL'] = sp.stats.randint(0, n_exp)
    return eud

#Impact function set
def _impfset_uncfunc(IFi, MDD, PAA, impf_set, haz_id_dict):
    impf_set_tmp = copy.deepcopy(impf_set)
    for haz_type, fun_id_list in haz_id_dict.items():
        for fun_id in fun_id_list:
            if MDD is not None:
                new_mdd = np.minimum(
                    impf_set_tmp.get_func(haz_type=haz_type, fun_id=fun_id).mdd * MDD,
                    1.0
                    )
                impf_set_tmp.get_func(haz_type=haz_type, fun_id=fun_id).mdd = new_mdd
            if PAA is not None:
                new_paa = np.minimum(
                    impf_set_tmp.get_func(haz_type=haz_type, fun_id=fun_id).paa * PAA,
                    1.0
                    )
                impf_set_tmp.get_func(haz_type=haz_type, fun_id=fun_id).paa = new_paa
            if IFi is not None:
                new_int = np.maximum(
                    impf_set_tmp.get_func(haz_type=haz_type, fun_id=fun_id).intensity + IFi,
                    0.0
                    )
                impf_set_tmp.get_func(haz_type=haz_type, fun_id=fun_id).intensity = new_int
    return impf_set_tmp

def _impfset_unc_dict(bounds_impfi, bounds_mdd, bounds_paa):
    iud = {}
    if bounds_impfi is not None:
        xmin, xdelta = bounds_impfi[0], bounds_impfi[1] - bounds_impfi[0]
        iud['IFi'] = sp.stats.uniform(xmin, xdelta)
    if bounds_paa is not None:
        xmin, xdelta = bounds_paa[0], bounds_paa[1] - bounds_paa[0]
        iud['PAA'] = sp.stats.uniform(xmin, xdelta)
    if bounds_mdd is not None:
        xmin, xdelta = bounds_mdd[0], bounds_mdd[1] - bounds_mdd[0]
        iud['MDD'] = sp.stats.uniform(xmin, xdelta)
    return iud

#Entity
def _disc_uncfunc(DR, disc_rate):
    disc = copy.deepcopy(disc_rate)
    if DR is not None:
        disc.rates = np.ones(disc.years.size) * DR
    return disc

def _disc_unc_dict(bounds_disk):
    if bounds_disk is None:
        return {}
    dmin, ddelta = bounds_disk[0], bounds_disk[1] - bounds_disk[0]
    return  {'DR': sp.stats.uniform(dmin, ddelta)}

def _meas_set_uncfunc(CO, meas_set):
    meas_set_tmp = copy.deepcopy(meas_set)
    if CO is not None:
        for haz_type in meas_set_tmp.get_hazard_types():
            for meas in meas_set_tmp.get_measure(haz_type=haz_type):
                meas.cost *= CO
    return meas_set_tmp

def _meas_set_unc_dict(bounds_cost):
    cmin, cdelta = bounds_cost[0], bounds_cost[1] - bounds_cost[0]
    return {'CO': sp.stats.uniform(cmin, cdelta)}

def _ent_unc_func(EN, ET, EL, IFi, MDD, PAA, CO, DR, bounds_noise,
                 impf_set, haz_id_dict, disc_rate, exp_list, meas_set):
    ent = Entity()
    ent.exposures = _exp_uncfunc(EN, ET, EL, exp_list, bounds_noise)
    ent.impact_funcs = _impfset_uncfunc(IFi, MDD, PAA, impf_set=impf_set,
                                            haz_id_dict=haz_id_dict)
    ent.measures = _meas_set_uncfunc(CO, meas_set=meas_set)
    ent.disc_rates = _disc_uncfunc(DR, disc_rate)
    return ent

def _ent_unc_dict(bounds_totval, bounds_noise, bounds_impfi, bounds_mdd,
                  bounds_paa, bounds_disc, bounds_cost, n_exp):
    ent_unc_dict = _exp_unc_dict(bounds_totval, bounds_noise, n_exp)
    ent_unc_dict.update(_impfset_unc_dict(bounds_impfi, bounds_mdd, bounds_paa))
    ent_unc_dict.update(_disc_unc_dict(bounds_disc))
    ent_unc_dict.update(_meas_set_unc_dict(bounds_cost))
    return  ent_unc_dict

def _entfut_unc_func(EN, EG, EL, IFi, MDD, PAA, CO, bounds_noise,
                 impf_set, haz_id_dict, exp_list, meas_set):
    ent = Entity()
    ent.exposures = _exp_uncfunc(EN=EN, ET=EG, EL=EL, exp_list=exp_list, bounds_noise=bounds_noise)
    ent.impact_funcs = _impfset_uncfunc(IFi, MDD, PAA, impf_set=impf_set,
                                            haz_id_dict=haz_id_dict)
    ent.measures = _meas_set_uncfunc(CO, meas_set=meas_set)
    ent.disc_rates = DiscRates() #Disc rate of future entity ignored in cost_benefit.calc()
    return ent

def _entfut_unc_dict(bounds_impfi, bounds_mdd,
                  bounds_paa, bounds_eg, bounds_noise,
                  bounds_cost, n_exp):
    eud = {}
    if bounds_eg is not None:
        gmin, gmax = bounds_eg[0], bounds_eg[1] - bounds_eg[0]
        eud['EG'] = sp.stats.uniform(gmin, gmax)
    if bounds_noise is not None:
        eud['EN'] = sp.stats.randint(0, 2**32 - 1) #seed for rnd generator
    if n_exp > 1:
        eud['EL'] = sp.stats.randint(0, n_exp)
    eud.update(_impfset_unc_dict(bounds_impfi, bounds_mdd, bounds_paa))
    if bounds_cost is not None:
        eud.update(_meas_set_unc_dict(bounds_cost))
    return eud



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Impact and ImpactFreqCurve classes.
"""

__all__ = ['ImpactFreqCurve', 'Impact']

import logging
import copy
import csv
import warnings
import datetime as dt
from itertools import zip_longest
import numpy as np
from scipy import sparse
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import pandas as pd
import xlsxwriter
from tqdm import tqdm


from climada.entity import Exposures, Tag
from climada.entity.exposures import INDICATOR_CENTR
from climada.hazard import Tag as TagHaz
import climada.util.plot as u_plot
from climada import CONFIG
from climada.util.constants import DEF_CRS, CMAP_IMPACT
import climada.util.coordinates as u_coord
import climada.util.dates_times as u_dt
from climada.util.select import get_attributes_with_matching_dimension

LOGGER = logging.getLogger(__name__)

class Impact():
    """Impact definition. Compute from an entity (exposures and impact
    functions) and hazard.

    Attributes
    ----------
    tag : dict
        dictionary of tags of exposures, impact functions set and
        hazard: {'exp': Tag(), 'impf_set': Tag(), 'haz': TagHazard()}
    event_id :
        np.array id (>0) of each hazard event
    event_name :
        list name of each hazard event
    date : np.array
        date if events as integer date corresponding to the
        proleptic Gregorian ordinal, where January 1 of year 1 has
        ordinal 1 (ordinal format of datetime library)
    coord_exp : np.array
        exposures coordinates [lat, lon] (in degrees)
    eai_exp : np.array
        expected annual impact for each exposure
    at_event : np.array
        impact for each hazard event
    frequency : np.array
        annual frequency of event
    tot_value : float
        total exposure value affected
    aai_agg : float
        average annual impact (aggregated)
    unit : str
        value unit used (given by exposures unit)
    imp_mat : sparse.csr_matrix
        matrix num_events x num_exp with impacts.
        only filled if save_mat is True in calc()
    """

    def __init__(self):
        """Empty initialization."""
        self.tag = dict()
        self.event_id = np.array([], int)
        self.event_name = list()
        self.date = np.array([], int)
        self.coord_exp = np.ndarray([], float)
        self.crs = DEF_CRS
        self.eai_exp = np.array([])
        self.at_event = np.array([])
        self.frequency = np.array([])
        self.tot_value = 0
        self.aai_agg = 0
        self.unit = ''
        self.imp_mat = sparse.csr_matrix(np.empty((0, 0)))

    def calc_freq_curve(self, return_per=None):
        """Compute impact exceedance frequency curve.

        Parameters
        ----------
        return_per : np.array, optional
            return periods where to compute
            the exceedance impact. Use impact's frequencies if not provided

        Returns
        -------
            ImpactFreqCurve
        """
        ifc = ImpactFreqCurve()
        ifc.tag = self.tag
        # Sort descendingly the impacts per events
        sort_idxs = np.argsort(self.at_event)[::-1]
        # Calculate exceedence frequency
        exceed_freq = np.cumsum(self.frequency[sort_idxs])
        # Set return period and imact exceeding frequency
        ifc.return_per = 1 / exceed_freq[::-1]
        ifc.impact = self.at_event[sort_idxs][::-1]
        ifc.unit = self.unit
        ifc.label = 'Exceedance frequency curve'

        if return_per is not None:
            interp_imp = np.interp(return_per, ifc.return_per, ifc.impact)
            ifc.return_per = return_per
            ifc.impact = interp_imp

        return ifc

    def calc(self, exposures, impact_funcs, hazard, save_mat=False):
        """Compute impact of an hazard to exposures.

        Parameters
        ----------
        exposures : climada.entity.Exposures
        impact_funcs : climada.entity.ImpactFuncSet
            impact functions
        hazard : climada.Hazard
        save_mat : bool
            self impact matrix: events x exposures

        Examples
        --------
            Use Entity class:

            >>> haz = Hazard.from_mat(HAZ_DEMO_MAT)  # Set hazard
            >>> haz.check()
            >>> ent = Entity.from_excel(ENT_TEMPLATE_XLS) # Set exposures
            >>> ent.check()
            >>> imp = Impact()
            >>> imp.calc(ent.exposures, ent.impact_funcs, haz)
            >>> imp.calc_freq_curve().plot()

            Specify only exposures and impact functions:

            >>> haz = Hazard.from_mat(HAZ_DEMO_MAT)  # Set hazard
            >>> haz.check()
            >>> funcs = ImpactFuncSet.from_excel(ENT_TEMPLATE_XLS) # Set impact functions
            >>> funcs.check()
            >>> exp = Exposures(pd.read_excel(ENT_TEMPLATE_XLS)) # Set exposures
            >>> exp.check()
            >>> imp = Impact()
            >>> imp.calc(exp, funcs, haz)
            >>> imp.aai_agg
        """
        # 1. Assign centroids to each exposure if not done
        assign_haz = INDICATOR_CENTR + hazard.tag.haz_type
        if assign_haz not in exposures.gdf:
            exposures.assign_centroids(hazard)
        else:
            LOGGER.info('Exposures matching centroids found in %s', assign_haz)

        # 2. Initialize values
        self.unit = exposures.value_unit
        self.event_id = hazard.event_id
        self.event_name = hazard.event_name
        self.date = hazard.date
        self.coord_exp = np.stack([exposures.gdf.latitude.values,
                                   exposures.gdf.longitude.values], axis=1)
        self.frequency = hazard.frequency
        self.at_event = np.zeros(hazard.intensity.shape[0])
        self.eai_exp = np.zeros(exposures.gdf.value.size)
        self.tag = {'exp': exposures.tag, 'impf_set': impact_funcs.tag,
                    'haz': hazard.tag}
        self.crs = exposures.crs

        # Select exposures with positive value and assigned centroid
        exp_idx = np.where((exposures.gdf.value > 0) & (exposures.gdf[assign_haz] >= 0))[0]
        if exp_idx.size == 0:
            LOGGER.warning("No affected exposures.")

        num_events = hazard.intensity.shape[0]
        LOGGER.info('Calculating damage for %s assets (>0) and %s events.',
                    exp_idx.size, num_events)

        # Get damage functions for this hazard
        impf_haz = exposures.get_impf_column(hazard.tag.haz_type)
        haz_imp = impact_funcs.get_func(hazard.tag.haz_type)

        # Check if deductible and cover should be applied
        insure_flag = False
        if ('deductible' in exposures.gdf) and ('cover' in exposures.gdf) \
        and exposures.gdf.cover.max():
            insure_flag = True

        if save_mat:
            # (data, (row_ind, col_ind))
            self.imp_mat = ([], ([], []))

        # 3. Loop over exposures according to their impact function
        tot_exp = 0
        for imp_fun in haz_imp:
            # get indices of all the exposures with this impact function
            exp_iimp = np.where(exposures.gdf[impf_haz].values[exp_idx] == imp_fun.id)[0]
            tot_exp += exp_iimp.size
            exp_step = CONFIG.max_matrix_size.int() // num_events
            if not exp_step:
                raise ValueError('Increase max_matrix_size configuration parameter to > %s'
                                 % str(num_events))
            # separate in chunks
            chk = -1
            for chk in range(int(exp_iimp.size / exp_step)):
                self._exp_impact(
                    exp_idx[exp_iimp[chk * exp_step:(chk + 1) * exp_step]],
                    exposures, hazard, imp_fun, insure_flag)
            self._exp_impact(exp_idx[exp_iimp[(chk + 1) * exp_step:]],
                             exposures, hazard, imp_fun, insure_flag)

        if not tot_exp:
            LOGGER.warning('No impact functions match the exposures.')
        self.aai_agg = sum(self.at_event * hazard.frequency)

        if save_mat:
            shape = (self.date.size, exposures.gdf.value.size)
            self.imp_mat = sparse.csr_matrix(self.imp_mat, shape=shape)

    def calc_risk_transfer(self, attachment, cover):
        """Compute traaditional risk transfer over impact. Returns new impact
        with risk transfer applied and the insurance layer resulting Impact metrics.

        Parameters
        ----------
        attachment : float
            (deductible)
        cover : float

        Returns
        -------
        climada.engine.Impact
        """
        new_imp = copy.deepcopy(self)
        if attachment or cover:
            imp_layer = np.minimum(np.maximum(new_imp.at_event - attachment, 0), cover)
            new_imp.at_event = np.maximum(new_imp.at_event - imp_layer, 0)
            new_imp.aai_agg = np.sum(new_imp.at_event * new_imp.frequency)
            # next values are no longer valid
            new_imp.eai_exp = np.array([])
            new_imp.coord_exp = np.array([])
            new_imp.imp_mat = sparse.csr_matrix(np.empty((0, 0)))
            # insurance layer metrics
            risk_transfer = copy.deepcopy(new_imp)
            risk_transfer.at_event = imp_layer
            risk_transfer.aai_agg = np.sum(imp_layer * new_imp.frequency)
            return new_imp, risk_transfer

        return new_imp, Impact()

    def plot_hexbin_eai_exposure(self, mask=None, ignore_zero=True,
                                 pop_name=True, buffer=0.0, extend='neither',
                                 axis=None, adapt_fontsize=True, **kwargs):
        """Plot hexbin expected annual impact of each exposure.

        Parameters
        ----------
            mask : np.array, optional
                mask to apply to eai_exp plotted.
            ignore_zero : bool, optional
                flag to indicate if zero and negative
                values are ignored in plot. Default: False
            pop_name : bool, optional
                add names of the populated places
            buffer : float, optional
                border to add to coordinates.
                Default: 1.0.
            extend : str, optional
                extend border colorbar with arrows.
                [ 'neither' | 'both' | 'min' | 'max' ]
            axis : matplotlib.axes._subplots.AxesSubplot, optional
                axis to use
            kwargs : optional
                arguments for hexbin matplotlib function

        Returns
        -------
            cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        if 'cmap' not in kwargs:
            kwargs['cmap'] = CMAP_IMPACT

        eai_exp = self._build_exp()
        axis = eai_exp.plot_hexbin(mask, ignore_zero, pop_name, buffer,
                                   extend, axis=axis, adapt_fontsize=adapt_fontsize, **kwargs)
        axis.set_title('Expected annual impact')
        return axis

    def plot_scatter_eai_exposure(self, mask=None, ignore_zero=True,
                                  pop_name=True, buffer=0.0, extend='neither',
                                  axis=None, adapt_fontsize=True, **kwargs):
        """Plot scatter expected annual impact of each exposure.

        Parameters
        ----------
        mask  : np.array, optional
            mask to apply to eai_exp plotted.
        ignore_zero : bool, optional
            flag to indicate if zero and negative
            values are ignored in plot. Default: False
        pop_name : bool, optional
            add names of the populated places
        buffer : float, optional
            border to add to coordinates.
            Default: 1.0.
        extend : str
            optional extend border colorbar with arrows.
            [ 'neither' | 'both' | 'min' | 'max' ]
        axis  : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        adapt_fontsize : bool, optional
                If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
                the default matplotlib font size is used. Default is True.
        kwargs : optional
            arguments for hexbin matplotlib function

        Returns
        -------
            cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        if 'cmap' not in kwargs:
            kwargs['cmap'] = CMAP_IMPACT

        eai_exp = self._build_exp()
        axis = eai_exp.plot_scatter(mask, ignore_zero, pop_name, buffer,
                                    extend, axis=axis, adapt_fontsize=adapt_fontsize, **kwargs)
        axis.set_title('Expected annual impact')
        return axis

    def plot_raster_eai_exposure(self, res=None, raster_res=None, save_tiff=None,
                                 raster_f=lambda x: np.log10((np.fmax(x + 1, 1))),
                                 label='value (log10)', axis=None, adapt_fontsize=True,
                                 **kwargs):
        """Plot raster expected annual impact of each exposure.

        Parameters
        ----------
        res : float, optional
            resolution of current data in units of latitude
            and longitude, approximated if not provided.
        raster_res : float, optional
            desired resolution of the raster
        save_tiff :  str, optional
            file name to save the raster in tiff
            format, if provided
        raster_f : lambda function
            transformation to use to data. Default: log10 adding 1.
        label : str colorbar label
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        adapt_fontsize : bool, optional
                If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
                the default matplotlib font size is used. Default is True.
        kwargs : optional
            arguments for imshow matplotlib function

        Returns
        -------
        cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        eai_exp = self._build_exp()
        axis = eai_exp.plot_raster(res, raster_res, save_tiff, raster_f,
                                   label, axis=axis, adapt_fontsize=adapt_fontsize, **kwargs)
        axis.set_title('Expected annual impact')
        return axis

    def plot_basemap_eai_exposure(self, mask=None, ignore_zero=False, pop_name=True,
                                  buffer=0.0, extend='neither', zoom=10,
                                  url='http://tile.stamen.com/terrain/tileZ/tileX/tileY.png',
                                  axis=None, **kwargs):
        """Plot basemap expected annual impact of each exposure.

        Parameters
        ----------
        mask : np.array, optional
            mask to apply to eai_exp plotted.
        ignore_zero : bool, optional
            flag to indicate if zero and negative
            values are ignored in plot. Default: False
        pop_name : bool, optional
            add names of the populated places
        buffer : float, optional
            border to add to coordinates. Default: 0.0.
        extend : str, optional
            extend border colorbar with arrows.
            [ 'neither' | 'both' | 'min' | 'max' ]
        zoom : int, optional
            zoom coefficient used in the satellite image
        url : str, optional
            image source, e.g. ctx.sources.OSM_C
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs : optional
            arguments for scatter matplotlib function, e.g.
            cmap='Greys'. Default: 'Wistia'

        Returns
        -------
        cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        if 'cmap' not in kwargs:
            kwargs['cmap'] = CMAP_IMPACT
        eai_exp = self._build_exp()
        axis = eai_exp.plot_basemap(mask, ignore_zero, pop_name, buffer,
                                    extend, zoom, url, axis=axis, **kwargs)
        axis.set_title('Expected annual impact')
        return axis

    def plot_hexbin_impact_exposure(self, event_id=1, mask=None, ignore_zero=True,
                                    pop_name=True, buffer=0.0, extend='neither',
                                    axis=None, adapt_fontsize=True, **kwargs):
        """Plot hexbin impact of an event at each exposure.
        Requires attribute imp_mat.

        Parameters
        ----------
        event_id : int, optional
            id of the event for which to plot the impact.
            Default: 1.
        mask : np.array, optional
            mask to apply to impact plotted.
        ignore_zero : bool, optional
            flag to indicate if zero and negative
            values are ignored in plot. Default: False
        pop_name : bool, optional
            add names of the populated places
        buffer : float, optional
            border to add to coordinates.
            Default: 1.0.
        extend : str, optional
            extend border colorbar with arrows.
            [ 'neither' | 'both' | 'min' | 'max' ]
        kwargs : optional
            arguments for hexbin matplotlib function
        axis : matplotlib.axes._subplots.AxesSubplot
            optional axis to use
        adapt_fontsize : bool, optional
            If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
            the default matplotlib font size is used. Default is True.

        Returns
        --------
            matplotlib.figure.Figure, cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        if not hasattr(self.imp_mat, "shape") or self.imp_mat.shape[1] == 0:
            raise ValueError('attribute imp_mat is empty. Recalculate Impact'
                             'instance with parameter save_mat=True')
        if 'cmap' not in kwargs:
            kwargs['cmap'] = CMAP_IMPACT
        impact_at_events_exp = self._build_exp_event(event_id)
        axis = impact_at_events_exp.plot_hexbin(mask, ignore_zero, pop_name,
                                                buffer, extend, axis=axis, adapt_fontsize=adapt_fontsize,
                                                **kwargs)

        return axis

    def plot_basemap_impact_exposure(self, event_id=1, mask=None, ignore_zero=True,
                                     pop_name=True, buffer=0.0, extend='neither', zoom=10,
                                     url='http://tile.stamen.com/terrain/tileZ/tileX/tileY.png',
                                     axis=None, **kwargs):
        """Plot basemap impact of an event at each exposure.
        Requires attribute imp_mat.

        Parameters
        ----------
        event_id : int, optional
            id of the event for which to plot the impact.
            Default: 1.
        mask  : np.array, optional
            mask to apply to impact plotted.
        ignore_zero : bool, optional
            flag to indicate if zero and negative
            values are ignored in plot. Default: False
        pop_name : bool, optional
            add names of the populated places
        buffer : float, optional
            border to add to coordinates. Default: 0.0.
        extend : str, optional
            extend border colorbar with arrows.
            [ 'neither' | 'both' | 'min' | 'max' ]
        zoom : int, optional
            zoom coefficient used in the satellite image
        url : str, optional
            image source, e.g. ctx.sources.OSM_C
        axis  : matplotlib.axes._subplots.AxesSubplot, optional axis to use
        kwargs : optional arguments for scatter matplotlib function, e.g.
            cmap='Greys'. Default: 'Wistia'

        Returns
        -------
        cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        if not hasattr(self.imp_mat, "shape") or self.imp_mat.shape[1] == 0:
            raise ValueError('attribute imp_mat is empty. Recalculate Impact'
                             'instance with parameter save_mat=True')

        if event_id not in self.event_id:
            raise ValueError(f'Event ID {event_id} not found')
        if 'cmap' not in kwargs:
            kwargs['cmap'] = CMAP_IMPACT
        impact_at_events_exp = self._build_exp_event(event_id)
        axis = impact_at_events_exp.plot_basemap(mask, ignore_zero, pop_name,
                                                 buffer, extend, zoom, url, axis=axis, **kwargs)

        return axis

    def write_csv(self, file_name):
        """Write data into csv file. imp_mat is not saved.

        Parameters
        ----------
        file_name : str
            absolute path of the file
        """
        LOGGER.info('Writing %s', file_name)
        with open(file_name, "w") as imp_file:
            imp_wr = csv.writer(imp_file)
            imp_wr.writerow(["tag_hazard", "tag_exposure", "tag_impact_func",
                             "unit", "tot_value", "aai_agg", "event_id",
                             "event_name", "event_date", "event_frequency",
                             "at_event", "eai_exp", "exp_lat", "exp_lon", "exp_crs"])
            csv_data = [[[self.tag['haz'].haz_type], [self.tag['haz'].file_name],
                         [self.tag['haz'].description]],
                        [[self.tag['exp'].file_name], [self.tag['exp'].description]],
                        [[self.tag['impf_set'].file_name], [self.tag['impf_set'].description]],
                        [self.unit], [self.tot_value], [self.aai_agg],
                        self.event_id, self.event_name, self.date,
                        self.frequency, self.at_event,
                        self.eai_exp, self.coord_exp[:, 0], self.coord_exp[:, 1],
                        [str(self.crs)]]
            for values in zip_longest(*csv_data):
                imp_wr.writerow(values)

    def write_excel(self, file_name):
        """Write data into Excel file. imp_mat is not saved.

        Parameters
        ----------
        file_name : str
            absolute path of the file
        """
        LOGGER.info('Writing %s', file_name)
        def write_col(i_col, imp_ws, xls_data):
            """Write one measure"""
            row_ini = 1
            for dat_row in xls_data:
                imp_ws.write(row_ini, i_col, dat_row)
                row_ini += 1

        imp_wb = xlsxwriter.Workbook(file_name)
        imp_ws = imp_wb.add_worksheet()

        header = ["tag_hazard", "tag_exposure", "tag_impact_func",
                  "unit", "tot_value", "aai_agg", "event_id",
                  "event_name", "event_date", "event_frequency",
                  "at_event", "eai_exp", "exp_lat", "exp_lon", "exp_crs"]
        for icol, head_dat in enumerate(header):
            imp_ws.write(0, icol, head_dat)
        data = [self.tag['haz'].haz_type, str(self.tag['haz'].file_name),
                str(self.tag['haz'].description)]
        write_col(0, imp_ws, data)
        data = [str(self.tag['exp'].file_name), str(self.tag['exp'].description)]
        write_col(1, imp_ws, data)
        data = [str(self.tag['impf_set'].file_name), str(self.tag['impf_set'].description)]
        write_col(2, imp_ws, data)
        write_col(3, imp_ws, [self.unit])
        write_col(4, imp_ws, [self.tot_value])
        write_col(5, imp_ws, [self.aai_agg])
        write_col(6, imp_ws, self.event_id)
        write_col(7, imp_ws, self.event_name)
        write_col(8, imp_ws, self.date)
        write_col(9, imp_ws, self.frequency)
        write_col(10, imp_ws, self.at_event)
        write_col(11, imp_ws, self.eai_exp)
        write_col(12, imp_ws, self.coord_exp[:, 0])
        write_col(13, imp_ws, self.coord_exp[:, 1])
        write_col(14, imp_ws, [str(self.crs)])

        imp_wb.close()

    def write_sparse_csr(self, file_name):
        """Write imp_mat matrix in numpy's npz format."""
        LOGGER.info('Writing %s', file_name)
        np.savez(file_name, data=self.imp_mat.data, indices=self.imp_mat.indices,
                 indptr=self.imp_mat.indptr, shape=self.imp_mat.shape)

    def calc_impact_year_set(self, all_years=True, year_range=None):
        """Calculate yearly impact from impact data.

        Parameters
        ----------
        all_years : boolean
            return values for all years between first and
            last year with event, including years without any events.
        year_range : tuple or list with integers
            start and end year

        Returns
        -------
        Impact year set of type numpy.ndarray with summed impact per year.
        """
        if year_range is None:
            year_range = []

        orig_year = np.array([dt.datetime.fromordinal(date).year
                              for date in self.date])
        if orig_year.size == 0 and len(year_range) == 0:
            return dict()
        if orig_year.size == 0 or (len(year_range) > 0 and all_years):
            years = np.arange(min(year_range), max(year_range) + 1)
        elif all_years:
            years = np.arange(min(orig_year), max(orig_year) + 1)
        else:
            years = np.array(sorted(np.unique(orig_year)))
        if not len(year_range) == 0:
            years = years[years >= min(year_range)]
            years = years[years <= max(year_range)]

        year_set = dict()

        for year in years:
            year_set[year] = sum(self.at_event[orig_year == year])
        return year_set

    def local_exceedance_imp(self, return_periods=(25, 50, 100, 250)):
        """Compute exceedance impact map for given return periods.
        Requires attribute imp_mat.

        Parameters
        ----------
        return_periods : np.array return periods to consider

        Returns
        -------
        np.array
        """
        LOGGER.info('Computing exceedance impact map for return periods: %s',
                    return_periods)
        try:
            self.imp_mat.shape[1]
        except AttributeError as err:
            raise ValueError('attribute imp_mat is empty. Recalculate Impact'
                             'instance with parameter save_mat=True') from err
        num_cen = self.imp_mat.shape[1]
        imp_stats = np.zeros((len(return_periods), num_cen))
        cen_step = CONFIG.max_matrix_size.int() // self.imp_mat.shape[0]
        if not cen_step:
            raise ValueError('Increase max_matrix_size configuration parameter to > %s'
                             % str(self.imp_mat.shape[0]))
        # separte in chunks
        chk = -1
        for chk in range(int(num_cen / cen_step)):
            self._loc_return_imp(np.array(return_periods),
                                 self.imp_mat[:, chk * cen_step:(chk + 1) * cen_step].toarray(),
                                 imp_stats[:, chk * cen_step:(chk + 1) * cen_step])
        self._loc_return_imp(np.array(return_periods),
                             self.imp_mat[:, (chk + 1) * cen_step:].toarray(),
                             imp_stats[:, (chk + 1) * cen_step:])

        return imp_stats

    def plot_rp_imp(self, return_periods=(25, 50, 100, 250),
                    log10_scale=True, smooth=True, axis=None, **kwargs):
        """Compute and plot exceedance impact maps for different return periods.
        Calls local_exceedance_imp.

        Parameters
        ----------
        return_periods : tuple(int), optional
            return periods to consider
        log10_scale : boolean, optional
            plot impact as log10(impact)
        smooth : bool, optional
            smooth plot to plot.RESOLUTIONxplot.RESOLUTION
        kwargs : optional
            arguments for pcolormesh matplotlib function
            used in event plots

        Returns
        --------
        matplotlib.axes._subplots.AxesSubplot,
        np.ndarray (return_periods.size x num_centroids)
        """
        imp_stats = self.local_exceedance_imp(np.array(return_periods))
        if imp_stats == []:
            raise ValueError('Error: Attribute imp_mat is empty. Recalculate Impact'
                             'instance with parameter save_mat=True')
        if log10_scale:
            if np.min(imp_stats) < 0:
                imp_stats_log = np.log10(abs(imp_stats) + 1)
                colbar_name = 'Log10(abs(Impact)+1) (' + self.unit + ')'
            elif np.min(imp_stats) < 1:
                imp_stats_log = np.log10(imp_stats + 1)
                colbar_name = 'Log10(Impact+1) (' + self.unit + ')'
            else:
                imp_stats_log = np.log10(imp_stats)
                colbar_name = 'Log10(Impact) (' + self.unit + ')'
        else:
            imp_stats_log = imp_stats
            colbar_name = 'Impact (' + self.unit + ')'
        title = list()
        for ret in return_periods:
            title.append('Return period: ' + str(ret) + ' years')
        axis = u_plot.geo_im_from_array(imp_stats_log, self.coord_exp,
                                        colbar_name, title, smooth=smooth, axes=axis, **kwargs)

        return axis, imp_stats

    @staticmethod
    def read_sparse_csr(file_name):
        """Read imp_mat matrix from numpy's npz format.

        Parameters
        ----------
        file_name : str file name

        Returns
        -------
        sparse.csr_matrix
        """
        LOGGER.info('Reading %s', file_name)
        loader = np.load(file_name)
        return sparse.csr_matrix(
            (loader['data'], loader['indices'], loader['indptr']), shape=loader['shape'])

    @classmethod
    def from_csv(cls, file_name):
        """Read csv file containing impact data generated by write_csv.

        Parameters
        ----------
        file_name : str absolute path of the file

        Returns
        -------
        imp : climada.engine.impact.Impact
            Impact from csv file
        """
        LOGGER.info('Reading %s', file_name)
        imp_df = pd.read_csv(file_name)
        imp = cls()
        imp.unit = imp_df.unit[0]
        imp.tot_value = imp_df.tot_value[0]
        imp.aai_agg = imp_df.aai_agg[0]
        imp.event_id = imp_df.event_id[~np.isnan(imp_df.event_id)].values
        num_ev = imp.event_id.size
        imp.event_name = imp_df.event_name[:num_ev].values.tolist()
        imp.date = imp_df.event_date[:num_ev].values
        imp.at_event = imp_df.at_event[:num_ev].values
        imp.frequency = imp_df.event_frequency[:num_ev].values
        imp.eai_exp = imp_df.eai_exp[~np.isnan(imp_df.eai_exp)].values
        num_exp = imp.eai_exp.size
        imp.coord_exp = np.zeros((num_exp, 2))
        imp.coord_exp[:, 0] = imp_df.exp_lat[:num_exp]
        imp.coord_exp[:, 1] = imp_df.exp_lon[:num_exp]
        try:
            imp.crs = u_coord.to_crs_user_input(imp_df.exp_crs.values[0])
        except AttributeError:
            imp.crs = DEF_CRS
        imp.tag['haz'] = TagHaz(str(imp_df.tag_hazard[0]),
                                 str(imp_df.tag_hazard[1]),
                                 str(imp_df.tag_hazard[2]))
        imp.tag['exp'] = Tag(str(imp_df.tag_exposure[0]),
                              str(imp_df.tag_exposure[1]))
        imp.tag['impf_set'] = Tag(str(imp_df.tag_impact_func[0]),
                                 str(imp_df.tag_impact_func[1]))
        return imp

    def read_csv(self, *args, **kwargs):
        """This function is deprecated, use Impact.from_csv instead."""
        LOGGER.warning("The use of Impact.read_csv is deprecated."
                       "Use Impact.from_csv instead.")
        self.__dict__ = Impact.from_csv(*args, **kwargs).__dict__

    @classmethod
    def from_excel(cls, file_name):
        """Read excel file containing impact data generated by write_excel.

        Parameters
        ----------
        file_name : str absolute path of the file

        Returns
        -------
        imp : climada.engine.impact.Impact
            Impact from excel file
        """
        LOGGER.info('Reading %s', file_name)
        dfr = pd.read_excel(file_name)
        imp =cls()
        imp.tag['haz'] = TagHaz()
        imp.tag['haz'].haz_type = dfr['tag_hazard'][0]
        imp.tag['haz'].file_name = dfr['tag_hazard'][1]
        imp.tag['haz'].description = dfr['tag_hazard'][2]
        imp.tag['exp'] = Tag()
        imp.tag['exp'].file_name = dfr['tag_exposure'][0]
        imp.tag['exp'].description = dfr['tag_exposure'][1]
        imp.tag['impf_set'] = Tag()
        imp.tag['impf_set'].file_name = dfr['tag_impact_func'][0]
        imp.tag['impf_set'].description = dfr['tag_impact_func'][1]

        imp.unit = dfr.unit[0]
        imp.tot_value = dfr.tot_value[0]
        imp.aai_agg = dfr.aai_agg[0]

        imp.event_id = dfr.event_id[~np.isnan(dfr.event_id.values)].values
        imp.event_name = dfr.event_name[:imp.event_id.size].values
        imp.date = dfr.event_date[:imp.event_id.size].values
        imp.frequency = dfr.event_frequency[:imp.event_id.size].values
        imp.at_event = dfr.at_event[:imp.event_id.size].values

        imp.eai_exp = dfr.eai_exp[~np.isnan(dfr.eai_exp.values)].values
        imp.coord_exp = np.zeros((imp.eai_exp.size, 2))
        imp.coord_exp[:, 0] = dfr.exp_lat.values[:imp.eai_exp.size]
        imp.coord_exp[:, 1] = dfr.exp_lon.values[:imp.eai_exp.size]
        try:
            imp.crs = u_coord.to_csr_user_input(dfr.exp_crs.values[0])
        except AttributeError:
            imp.crs = DEF_CRS

        return imp

    def read_excel(self, *args, **kwargs):
        """This function is deprecated, use Impact.from_excel instead."""
        LOGGER.warning("The use of Impact.read_excel is deprecated."
                       "Use Impact.from_excel instead.")
        self.__dict__ = Impact.from_excel(*args, **kwargs).__dict__

    @staticmethod
    def video_direct_impact(exp, impf_set, haz_list, file_name='',
                            writer=animation.PillowWriter(bitrate=500),
                            imp_thresh=0, args_exp=None, args_imp=None):
        """
        Computes and generates video of accumulated impact per input events
        over exposure.

        Parameters
        ----------
        exp : Exposures
            exposures instance, constant during all video
        impf_set : ImpactFuncSet
            impact functions
        haz_list : (list(Hazard))
            every Hazard contains an event; all hazards
            use the same centroids
        file_name : str, optional
            file name to save video, if provided
        writer : matplotlib.animation.*, optional
            video writer. Default: pillow with bitrate=500
        imp_thresh : float
            represent damages greater than threshold
        args_exp : optional
            arguments for scatter (points) or hexbin (raster)
            matplotlib function used in exposures
        args_imp : optional
            arguments for scatter (points) or hexbin (raster)
            matplotlib function used in impact

        Returns
        -------
        list(Impact)
        """
        if args_exp is None:
            args_exp = dict()
        if args_imp is None:
            args_imp = dict()
        imp_list = []
        exp_list = []
        imp_arr = np.zeros(len(exp.gdf))
        for i_time, _ in enumerate(haz_list):
            imp_tmp = Impact()
            imp_tmp.calc(exp, impf_set, haz_list[i_time])
            imp_arr = np.maximum(imp_arr, imp_tmp.eai_exp)
            # remove not impacted exposures
            save_exp = imp_arr > imp_thresh
            imp_tmp.coord_exp = imp_tmp.coord_exp[save_exp, :]
            imp_tmp.eai_exp = imp_arr[save_exp]
            imp_list.append(imp_tmp)
            exp_list.append(~save_exp)

        v_lim = [np.array([haz.intensity.min() for haz in haz_list]).min(),
                 np.array([haz.intensity.max() for haz in haz_list]).max()]

        if 'vmin' not in args_exp:
            args_exp['vmin'] = exp.gdf.value.values.min()

        if 'vmin' not in args_imp:
            args_imp['vmin'] = np.array([imp.eai_exp.min() for imp in imp_list
                                         if imp.eai_exp.size]).min()

        if 'vmax' not in args_exp:
            args_exp['vmax'] = exp.gdf.value.values.max()

        if 'vmax' not in args_imp:
            args_imp['vmax'] = np.array([imp.eai_exp.max() for imp in imp_list
                                         if imp.eai_exp.size]).max()

        if 'cmap' not in args_exp:
            args_exp['cmap'] = 'winter_r'

        if 'cmap' not in args_imp:
            args_imp['cmap'] = 'autumn_r'


        plot_raster = False
        if exp.meta:
            plot_raster = True

        def run(i_time):
            haz_list[i_time].plot_intensity(1, axis=axis, cmap='Greys', vmin=v_lim[0],
                                            vmax=v_lim[1], alpha=0.8)
            if plot_raster:
                exp.plot_hexbin(axis=axis, mask=exp_list[i_time], ignore_zero=True,
                                pop_name=False, **args_exp)
                if imp_list[i_time].coord_exp.size:
                    imp_list[i_time].plot_hexbin_eai_exposure(axis=axis, pop_name=False,
                                                              **args_imp)
                    fig.delaxes(fig.axes[1])
            else:
                exp.plot_scatter(axis=axis, mask=exp_list[i_time], ignore_zero=True,
                                 pop_name=False, **args_exp)
                if imp_list[i_time].coord_exp.size:
                    imp_list[i_time].plot_scatter_eai_exposure(axis=axis, pop_name=False,
                                                               **args_imp)
                    fig.delaxes(fig.axes[1])
            fig.delaxes(fig.axes[1])
            fig.delaxes(fig.axes[1])
            axis.set_xlim(haz_list[-1].centroids.lon.min(), haz_list[-1].centroids.lon.max())
            axis.set_ylim(haz_list[-1].centroids.lat.min(), haz_list[-1].centroids.lat.max())
            axis.set_title(haz_list[i_time].event_name[0])
            pbar.update()

        if file_name:
            LOGGER.info('Generating video %s', file_name)
            fig, axis, _fontsize = u_plot.make_map()
            ani = animation.FuncAnimation(fig, run, frames=len(haz_list),
                                          interval=500, blit=False)
            pbar = tqdm(total=len(haz_list))
            fig.tight_layout()
            ani.save(file_name, writer=writer)
            pbar.close()

        return imp_list

    def _loc_return_imp(self, return_periods, imp, exc_imp):
        """Compute local exceedence impact for given return period.

        Parameters
        ----------
        return_periods : np.array return periods to consider
        cen_pos (int): centroid position

        Returns
        -------
        np.array
        """
        # sorted impacts
        sort_pos = np.argsort(imp, axis=0)[::-1, :]
        columns = np.ones(imp.shape, int)
        # pylint: disable=unsubscriptable-object  # pylint/issues/3139
        columns *= np.arange(columns.shape[1])
        imp_sort = imp[sort_pos, columns]
        # cummulative frequency at sorted intensity
        freq_sort = self.frequency[sort_pos]
        np.cumsum(freq_sort, axis=0, out=freq_sort)

        for cen_idx in range(imp.shape[1]):
            exc_imp[:, cen_idx] = self._cen_return_imp(
                imp_sort[:, cen_idx], freq_sort[:, cen_idx],
                0, return_periods)

    def _exp_impact(self, exp_iimp, exposures, hazard, imp_fun, insure_flag):
        """Compute impact for inpute exposure indexes and impact function.

        Parameters
        ----------
        exp_iimp : np.array exposures indexes
        exposures: climada.entity.Exposures instance
        hazard : climada.Hazard
        imp_fun : climada.entity.ImpactFunc
            impact function instance
        insure_flag : bool
            consider deductible and cover of exposures
        """
        if not exp_iimp.size:
            return

        # get assigned centroids
        icens = exposures.gdf[INDICATOR_CENTR + hazard.tag.haz_type].values[exp_iimp]

        # get affected intensities
        inten_val = hazard.intensity[:, icens]
        # get affected fractions
        fract = hazard.fraction[:, icens]
        # impact = fraction * mdr * value
        inten_val.data = imp_fun.calc_mdr(inten_val.data)
        impact = fract.multiply(inten_val).multiply(exposures.gdf.value.values[exp_iimp])

        if insure_flag and impact.nonzero()[0].size:
            inten_val = hazard.intensity[:, icens].toarray()
            paa = np.interp(inten_val, imp_fun.intensity, imp_fun.paa)
            impact = impact.toarray()
            impact -= exposures.gdf.deductible.values[exp_iimp] * paa
            impact = np.clip(impact, 0, exposures.gdf.cover.values[exp_iimp])
            self.eai_exp[exp_iimp] += np.einsum('ji,j->i', impact, hazard.frequency)
            impact = sparse.coo_matrix(impact)
        else:
            self.eai_exp[exp_iimp] += np.squeeze(np.asarray(np.sum(
                impact.multiply(hazard.frequency.reshape(-1, 1)), axis=0)))

        self.at_event += np.squeeze(np.asarray(np.sum(impact, axis=1)))
        self.tot_value += np.sum(exposures.gdf.value.values[exp_iimp])
        if isinstance(self.imp_mat, tuple):
            row_ind, col_ind = impact.nonzero()
            self.imp_mat[0].extend(list(impact.data))
            self.imp_mat[1][0].extend(list(row_ind))
            self.imp_mat[1][1].extend(list(exp_iimp[col_ind]))

    def _build_exp(self):
        return Exposures(
            data={
                'value': self.eai_exp,
                'latitude': self.coord_exp[:, 0],
                'longitude': self.coord_exp[:, 1],
            },
            crs=self.crs,
            value_unit=self.unit,
            ref_year=0,
            tag=Tag(),
            meta=None
        )

    def _build_exp_event(self, event_id):
        """Write impact of an event as Exposures

        Parameters
        ----------
        event_id : int
            id of the event
        """
        [[ix]] = (self.event_id == event_id).nonzero()
        return Exposures(
            data={
                'value': self.imp_mat[ix].toarray().ravel(),
                'latitude': self.coord_exp[:, 0],
                'longitude': self.coord_exp[:, 1],
            },
            crs=self.crs,
            value_unit=self.unit,
            ref_year=0,
            tag=Tag(),
            meta=None
        )

    @staticmethod
    def _cen_return_imp(imp, freq, imp_th, return_periods):
        """From ordered impact and cummulative frequency at centroid, get
        exceedance impact at input return periods.

        Parameters
        ----------
        imp : np.array
            sorted impact at centroid
        freq : np.array
            cummulative frequency at centroid
        imp_th : float
            impact threshold
        return_periods : np.array
            return periods

        Returns
        -------
        np.array
        """
        imp_th = np.asarray(imp > imp_th).squeeze()
        imp_cen = imp[imp_th]
        freq_cen = freq[imp_th]
        if not imp_cen.size:
            return np.zeros((return_periods.size,))
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                pol_coef = np.polyfit(np.log(freq_cen), imp_cen, deg=1)
        except ValueError:
            pol_coef = np.polyfit(np.log(freq_cen), imp_cen, deg=0)
        imp_fit = np.polyval(pol_coef, np.log(1 / return_periods))
        wrong_inten = (return_periods > np.max(1 / freq_cen)) & np.isnan(imp_fit)
        imp_fit[wrong_inten] = 0.

        return imp_fit


    def select(self,
               event_ids=None, event_names=None, dates=None,
               coord_exp=None):
        """
        Select a subset of events and/or exposure points from the impact.
        If multiple input variables are not None, it returns all the impacts
        matching at least one of the conditions.

        Note
        ----
            the frequencies are NOT adjusted. Method to adjust frequencies
        and obtain correct eai_exp:
            1- Select subset of impact according to your choice
            imp = impact.select(...)
            2- Adjust manually the frequency of the subset of impact
            imp.frequency = [...]
            3- Use select without arguments to select all events and recompute
            the eai_exp with the updated frequencies.
            imp = imp.select()

        Parameters
        ----------
        event_ids : list[int], optional
            Selection of events by their id. The default is None.
        event_names : list[str], optional
            Selection of events by their name. The default is None.
        dates : tuple(), optional
            (start-date, end-date), events are selected if they are >=
            than start-date and <= than end-date. Dates in same format
            as impact.date (ordinal format of datetime library)
            The default is None.
        coord_exp : np.ndarray), optional
            Selection of exposures coordinates [lat, lon] (in degrees)
            The default is None.

        Raises
        ------
        ValueError
            If the impact matrix is missing, the eai_exp and aai_agg cannot
            be updated for a selection of events and/or exposures.

        Returns
        -------
        imp : climada.engine.Impact
            A new impact object with a selection of events and/or exposures

        """

        nb_events = self.event_id.size
        nb_exp = len(self.coord_exp)

        if self.imp_mat.shape != (nb_events, nb_exp):
            raise ValueError("The impact matrix is missing or incomplete. "
                             "The eai_exp and aai_agg cannot be computed. "
                             "Please recompute impact.calc() with save_mat=True "
                             "before using impact.select()")

        if nb_events == nb_exp:
            LOGGER.warning("The number of events is equal to the number of "
                           "exposure points. It is not possible to "
                           "differentiate events and exposures attributes. "
                           "Please add/remove one event/exposure point. "
                           "This is a purely technical limitation of this "
                           "method.")
            return None

        imp = copy.deepcopy(self)

        # apply event selection to impact attributes
        sel_ev = self._selected_events_idx(event_ids, event_names, dates, nb_events)
        if sel_ev is not None:
            # set all attributes that are 'per event', i.e. have a dimension
            # of length equal to the number of events (=nb_events)
            for attr in get_attributes_with_matching_dimension(imp, [nb_events]):
                value = imp.__getattribute__(attr)
                if isinstance(value, np.ndarray):
                    if value.ndim == 1:
                        setattr(imp, attr, value[sel_ev])
                    else:
                        LOGGER.warning("Found a multidimensional numpy array "
                                       "with one dimension matching the number of events. "
                                       "But multidimensional numpy arrays are not handled "
                                       "in impact.select")
                elif isinstance(value, sparse.csr_matrix):
                    setattr(imp, attr, value[sel_ev, :])
                elif isinstance(value, list) and value:
                    setattr(imp, attr, [value[idx] for idx in sel_ev])
                else:
                    pass

            LOGGER.info("The eai_exp and aai_agg are computed for the "
                        "selected subset of events WITHOUT modification of "
                        "the frequencies.")

        # apply exposure selection to impact attributes
        if coord_exp is not None:
            sel_exp = self._selected_exposures_idx(coord_exp)
            imp.coord_exp = imp.coord_exp[sel_exp]
            imp.imp_mat = imp.imp_mat[:, sel_exp]

            # .A1 reduce 1d matrix to 1d array
            imp.at_event = imp.imp_mat.sum(axis=1).A1
            imp.tot_value = None
            LOGGER.info("The total value cannot be re-computed for a "
                        "subset of exposures and is set to None.")

        # cast frequency vector into 2d array for sparse matrix multiplication
        freq_mat = imp.frequency.reshape(len(imp.frequency), 1)
        # .A1 reduce 1d matrix to 1d array
        imp.eai_exp = imp.imp_mat.multiply(freq_mat).sum(axis=0).A1
        imp.aai_agg = imp.eai_exp.sum()

        return imp

    def _selected_exposures_idx(self, coord_exp):
        assigned_idx = u_coord.assign_coordinates(self.coord_exp, coord_exp, threshold=0)
        sel_exp = (assigned_idx >= 0).nonzero()[0]
        if sel_exp.size == 0:
            LOGGER.warning("No exposure coordinates match the selection.")
        return sel_exp

    def _selected_events_idx(self, event_ids, event_names, dates, nb_events):
        if all(var is None for var in [dates, event_ids, event_names]):
            return None

        # filter events by date
        if dates is None:
            mask_dt = np.zeros(nb_events, dtype=bool)
        else:
            mask_dt = np.ones(nb_events, dtype=bool)
            date_ini, date_end = dates
            if isinstance(date_ini, str):
                date_ini = u_dt.str_to_date(date_ini)
                date_end = u_dt.str_to_date(date_end)
            mask_dt &= (date_ini <= self.date)
            mask_dt &= (self.date <= date_end)
            if not np.any(mask_dt):
                LOGGER.info('No impact event in given date range %s.', dates)

        sel_dt = mask_dt.nonzero()[0]  # Convert bool to indices

        # filter events by id
        if event_ids is None:
            sel_id = np.array([], dtype=int)
        else:
            sel_id = np.isin(self.event_id, event_ids).nonzero()[0]
            if sel_id.size == 0:
                LOGGER.info('No impact event with given ids %s found.', event_ids)

        # filter events by name
        if event_names is None:
            sel_na = np.array([], dtype=int)
        else:
            sel_na = np.isin(self.event_name, event_names).nonzero()[0]
            if sel_na.size == 0:
                LOGGER.info('No impact event with given names %s found.', event_names)

        # select events with machting id, name or date field.
        sel_ev = np.unique(np.concatenate([sel_dt, sel_id, sel_na]))

        # if no event found matching ids, names or dates, warn the user
        if sel_ev.size == 0:
            LOGGER.warning("No event matches the selection. ")

        return sel_ev

class ImpactFreqCurve():
    """Impact exceedence frequency curve.

    Attributes
    ----------
    tag : dict
        dictionary of tags of exposures, impact functions set and
        hazard: {'exp': Tag(), 'impf_set': Tag(), 'haz': TagHazard()}
    return_per : np.array
        return period
    impact : np.array
        impact exceeding frequency
    unit : str
        value unit used (given by exposures unit)
    label : str
        string describing source data
    """
    def __init__(self):
        self.tag = dict()
        self.return_per = np.array([])
        self.impact = np.array([])
        self.unit = ''
        self.label = ''

    def plot(self, axis=None, log_frequency=False, **kwargs):
        """Plot impact frequency curve.

        Parameters
        ----------
        axis  : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        log_frequency : boolean, optional
            plot logarithmioc exceedance frequency on x-axis
        kwargs : optional
            arguments for plot matplotlib function, e.g. color='b'

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        if not axis:
            _, axis = plt.subplots(1, 1)
        axis.set_title(self.label)
        axis.set_ylabel('Impact (' + self.unit + ')')
        if log_frequency:
            axis.set_xlabel('Exceedance frequency (1/year)')
            axis.set_xscale('log')
            axis.plot(self.return_per**-1, self.impact, **kwargs)
        else:
            axis.set_xlabel('Return period (year)')
            axis.plot(self.return_per, self.impact, **kwargs)
        return axis



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Impact class.
"""
import unittest
import numpy as np

from climada import CONFIG
from climada.util.constants import DEMO_DIR

import climada.engine.impact_data as im_d

DATA_DIR = CONFIG.engine.test_data.dir()
EMDAT_TEST_CSV = DATA_DIR.joinpath('emdat_testdata_BGD_USA_1970-2017.csv')
EMDAT_TEST_CSV_FAKE = DATA_DIR.joinpath('emdat_testdata_fake_2007-2011.csv')
EMDAT_2020_CSV_DEMO = DEMO_DIR.joinpath('demo_emdat_impact_data_2020.csv')

class TestEmdatImport(unittest.TestCase):
    """Test import of EM-DAT data (as CSV) for impact data analysis"""

    def test_clean_emdat_df_2018_load(self):
        """load selected sub sample from CSV, return DataFrame.
            here: from 2018 EM-DAT version to 2018 target_version"""

        df = im_d.clean_emdat_df(EMDAT_TEST_CSV, countries=['Bangladesh'], hazard='TC',
                                 year_range=[2000, 2017], target_version=2018)
        self.assertIn('ISO', df.columns)
        self.assertIn('Year', df.columns)
        iso3 = list(df.ISO.unique())
        years = np.arange(df.Year.min(), df.Year.max() + 1)

        self.assertListEqual(['BGD'], iso3)
        self.assertEqual(18, len(years))
        self.assertEqual(2017, years[-1])
        self.assertEqual(2010, years[10])
        self.assertEqual(450, df.size)
        self.assertEqual(8978541, df['Total affected'].max())
        self.assertIn('Tropical cyclone', list(df['Disaster subtype']))
        self.assertFalse(False in list(df['Disaster subtype'] == 'Tropical cyclone'))
        self.assertFalse('Flood' in list(df['Disaster subtype']))

    def test_emdat_df_2018_to_2020_load(self):
        """load selected sub sample from CSV, return DataFrame
                here: from 2018 EM-DAT version to 2020 target_version"""
        df = im_d.clean_emdat_df(EMDAT_TEST_CSV, countries=['USA'], hazard='TC',
                                 year_range=[2000, 2017], target_version=2020)
        self.assertIn('ISO', df.columns)
        self.assertIn('Year', df.columns)
        iso3 = list(df.ISO.unique())
        years = np.arange(df.Year.min(), df.Year.max() + 1)
        self.assertListEqual(['USA'], iso3)
        self.assertEqual(18, len(years))
        self.assertEqual(2017, years[-1])
        self.assertEqual(2010, years[10])
        self.assertEqual(1634, df.size)
        self.assertEqual(60000000, df["Insured Damages ('000 US$)"].max())
        self.assertIn('Tropical cyclone', list(df['Disaster Subtype']))
        self.assertFalse(False in list(df['Disaster Subtype'] == 'Tropical cyclone'))
        self.assertFalse('Flood' in list(df['Disaster Subtype']))

    def test_emdat_df_2020_load(self):
        """load selected sub sample from CSV, return DataFrame
                here: from 2020 EM-DAT version to 2020 target_version"""
        df = im_d.clean_emdat_df(EMDAT_2020_CSV_DEMO, countries=['THA', 'Viet Nam'], hazard='TC',
                                 year_range=[2005, 2008], target_version=2020)
        self.assertIn('ISO', df.columns)
        self.assertIn('Year', df.columns)
        iso3 = list(df.ISO.unique())
        years = np.arange(df.Year.min(), df.Year.max() + 1)
        self.assertIn('THA', iso3)
        self.assertIn('VNM', iso3)
        self.assertNotIn('USA', iso3)
        self.assertNotIn('TWN', iso3)
        self.assertEqual(4, len(years))
        self.assertEqual(2008, years[-1])
        self.assertEqual(2006, years[1])
        self.assertEqual(43, df.columns.size)
        self.assertEqual(688, df.size)
        self.assertEqual(624000, df["Total Damages ('000 US$)"].max())
        self.assertIn('Tropical cyclone', list(df['Disaster Subtype']))
        self.assertFalse(False in list(df['Disaster Subtype'] == 'Tropical cyclone'))
        self.assertFalse('Flood' in list(df['Disaster Subtype']))

class TestGDPScaling(unittest.TestCase):
    """test scaling of impact values proportional to GDP"""
    def test_scale_impact2refyear(self):
        """scale of impact values proportional to GDP"""
        impact_scaled = im_d.scale_impact2refyear([10, 100, 1000, 100, 100],
                                                  [1999, 2005, 2015, 2000, 2000],
                                                  ['CZE', 'CZE', 'MEX', 'MEX', 'CZE'],
                                                  reference_year=2015)
        # scaled impact value might change if worldbank input data changes,
        # check magnitude and adjust if test fails in the following line:
        self.assertListEqual(impact_scaled, [28, 137, 1000, 165, 304])

class TestEmdatProcessing(unittest.TestCase):
    def test_emdat_impact_event_2018(self):
        """test emdat_impact_event event impact data extraction, version 2018"""
        df = im_d.emdat_impact_event(EMDAT_TEST_CSV, countries=['Bangladesh', 'USA'],
                                     hazard='Drought', year_range=[2015, 2017],
                                     reference_year=2017, version=2018)

        self.assertEqual(46, df.size)
        self.assertEqual('2017-9550', df['Disaster No.'][1])
        self.assertEqual(df["Total damage ('000 US$)"][0],
                         df["impact"][0] * 1e-3)
        self.assertEqual(df["impact_scaled"][1],
                         df["impact"][1])
        self.assertEqual(df["Total damage ('000 US$)"][1], 2500000)
        self.assertEqual(df["Total damage ('000 US$)"][0], 1800000)
        # scaled impact value might change if worldbank input data changes,
        # check magnitude and adjust if test failes in the following 1 lines:
        self.assertAlmostEqual(df["impact_scaled"][0] * 1e-7,
                               192.7868, places=0)
        self.assertIn('USA', list(df['ISO']))
        self.assertIn('Drought', list(df['Disaster type']))
        self.assertEqual(2017, df['reference_year'].min())

    def test_emdat_impact_event_2020(self):
        """test emdat_impact_event event impact data extraction, version 2020"""
        df = im_d.emdat_impact_event(EMDAT_TEST_CSV, countries=['Bangladesh', 'USA'],
                                     hazard='Drought', year_range=[2015, 2017],
                                     reference_year=2000, version=2020)

        self.assertEqual(96, df.size)
        self.assertEqual('2017-9550', df['Dis No'][1])
        self.assertEqual(df["Total Damages ('000 US$)"][0],
                         df["impact"][0] * 1e-3)
        self.assertNotEqual(df["impact_scaled"][1],
                            df["impact"][1])
        self.assertEqual(df["Total Damages ('000 US$)"][1], 2500000)
        self.assertEqual(df["Total Damages ('000 US$)"][0], 1800000)
        # scaled impact value might change if worldbank input data changes,
        # check magnitude and adjust if test failes in the following line:
        self.assertAlmostEqual(df["impact_scaled"][0] * 1e-9,
                               1.012, places=0)
        self.assertIn('USA', list(df['ISO']))
        self.assertIn('Drought', list(df['Disaster Type']))
        self.assertEqual(2000, df['reference_year'].min())

    def test_emdat_affected_yearlysum(self):
        """test emdat_impact_yearlysum yearly impact data extraction"""
        df = im_d.emdat_impact_yearlysum(EMDAT_TEST_CSV, countries=['Bangladesh', 'USA'],
                                         hazard='Flood', year_range=(2015, 2017),
                                         reference_year=None, imp_str="Total Affected")

        self.assertEqual(36, df.size)
        self.assertEqual(df["impact"][1], 91000)
        self.assertEqual(df.impact.sum(), 11517946)
        self.assertEqual(df["year"][5], 2017)

        self.assertIn('USA', list(df['ISO']))
        self.assertIn('BGD', list(df['ISO']))

    def test_emdat_damage_yearlysum(self):
        """test emdat_impact_yearlysum yearly impact data extraction with scaling"""
        df = im_d.emdat_impact_yearlysum(EMDAT_TEST_CSV, countries=['Bangladesh', 'USA'],
                                         hazard='Flood', year_range=(2015, 2017),
                                         reference_year=2000)

        self.assertEqual(36, df.size)
        self.assertAlmostEqual(df.impact.max(), 15150000000.0)
        self.assertEqual(df.impact_scaled.min(), 10943000.0)
        self.assertEqual(df["year"][5], 2017)
        self.assertEqual(df["reference_year"].max(), 2000)
        self.assertIn('USA', list(df['ISO']))
        self.assertIn(50, list(df['region_id']))

    def test_emdat_countries_by_hazard_2020_pass(self):
        """test to get list of countries impacted by tropical cyclones from 2000 to 2019"""
        iso3_codes, country_names = im_d.emdat_countries_by_hazard(EMDAT_2020_CSV_DEMO,
                                                                   hazard='TC',
                                                                   year_range=(2000, 2019))

        self.assertIn('Réunion', country_names)
        self.assertEqual('Sri Lanka', country_names[4])
        self.assertEqual('BLZ', iso3_codes[3])
        self.assertEqual(len(country_names), len(iso3_codes))
        self.assertEqual(100, len(iso3_codes))

class TestEmdatToImpact(unittest.TestCase):
    """Test import of EM-DAT data (as CSV) to Impact-instance (CLIMADA)"""
    def test_emdat_to_impact_all_countries_pass(self):
        """test import EM-DAT to Impact() for all countries in CSV"""
        # =====================================================================
        # emdat_to_impact(emdat_file_csv, hazard_type_climada, \
        #                 year_range=None, countries=None, hazard_type_emdat=None, \
        #                 reference_year=None, imp_str="Total Damages ('000 US$)")
        # =====================================================================

        # file 1: version 2020
        _impact_emdat2020, countries2020 = im_d.emdat_to_impact(EMDAT_2020_CSV_DEMO, 'TC')
        # file 2: version 2018
        impact_emdat, countries = im_d.emdat_to_impact(EMDAT_TEST_CSV, 'TC')

        self.assertEqual(142, impact_emdat.event_id.size)
        self.assertEqual(141, impact_emdat.event_id[-1])
        self.assertEqual(0, impact_emdat.event_id[0])
        self.assertIn('2013-0138', impact_emdat.event_name)
        self.assertEqual('USA', countries[0])
        self.assertEqual('BGD', countries[1])
        self.assertEqual(len(countries), len(impact_emdat.eai_exp))
        self.assertEqual(2, len(impact_emdat.eai_exp))
        self.assertEqual(impact_emdat.date.size, impact_emdat.frequency.size)
        self.assertAlmostEqual(555861710000 * 1e-5, np.sum(impact_emdat.at_event) * 1e-5, places=0)
        self.assertAlmostEqual(0.0208333333333, np.unique(impact_emdat.frequency)[0], places=7)
        self.assertAlmostEqual(11580452291.666666, impact_emdat.aai_agg, places=0)
        self.assertAlmostEqual(109456249.99999999, impact_emdat.eai_exp[1], places=0)
        self.assertAlmostEqual(11470996041.666666, impact_emdat.eai_exp[0], places=0)
        self.assertIn('SPI', countries2020)
        self.assertNotIn('SPI', countries)

    def test_emdat_to_impact_scale(self):
        """test import DR EM-DAT to Impact() for 1 country and ref.year (scaling)"""
        impact_emdat = im_d.emdat_to_impact(EMDAT_TEST_CSV, 'DR',
                                            year_range=[2010, 2016], countries=['USA'],
                                            hazard_type_emdat='Drought',
                                            reference_year=2016)[0]
        self.assertEqual(5, impact_emdat.event_id.size)
        self.assertEqual(4, impact_emdat.event_id[-1])
        self.assertEqual(0, impact_emdat.event_id[0])
        self.assertIn('2012-9235', impact_emdat.event_name)
        self.assertEqual(1, len(impact_emdat.eai_exp))
        self.assertAlmostEqual(impact_emdat.aai_agg, impact_emdat.eai_exp[0])
        self.assertAlmostEqual(0.14285714, np.unique(impact_emdat.frequency)[0], places=3)
        # scaled impact value might change if worldbank input data changes,
        # check magnitude and adjust if test failes in the following 2 lines:
        self.assertAlmostEqual(3.69, np.sum(impact_emdat.at_event * 1e-10), places=0)
        self.assertAlmostEqual(5.28, impact_emdat.aai_agg * 1e-9, places=0)

    def test_emdat_to_impact_fakedata(self):
        """test import TC EM-DAT to Impact() for all countries in CSV"""
        impact_emdat, countries = im_d.emdat_to_impact(EMDAT_TEST_CSV_FAKE, 'FL',
                                                       hazard_type_emdat='Flood')
        self.assertEqual(6, impact_emdat.event_id.size)
        self.assertEqual(5, impact_emdat.event_id[-1])
        self.assertEqual(0, impact_emdat.event_id[0])
        self.assertIn('2008-0001', impact_emdat.event_name)
        self.assertEqual('CHE', countries[0])
        self.assertEqual('DEU', countries[1])
        self.assertEqual(len(countries), len(impact_emdat.eai_exp))
        self.assertEqual(2, len(impact_emdat.eai_exp))
        self.assertAlmostEqual(11000000.0, np.sum(impact_emdat.at_event))
        self.assertAlmostEqual(0.2, np.unique(impact_emdat.frequency)[0])
        self.assertAlmostEqual(2200000.0, impact_emdat.aai_agg)
        self.assertAlmostEqual(200000.0, impact_emdat.eai_exp[1])  # DEU
        self.assertAlmostEqual(2000000.0, impact_emdat.eai_exp[0])  # CHE

    def test_emdat_to_impact_2020format(self):
        """test import TC EM-DAT to Impact() from new 2020 EMDAT format CSV"""
        df1 = im_d.clean_emdat_df(EMDAT_2020_CSV_DEMO, hazard='TC',
                                  countries='PHL', year_range=(2013, 2013))
        df2 = im_d.emdat_impact_event(EMDAT_2020_CSV_DEMO, countries='PHL', hazard='TC',
                                      year_range=(2013, 2013), reference_year=None,
                                      imp_str='Total Affected')
        impact_emdat, _countries = im_d.emdat_to_impact(EMDAT_2020_CSV_DEMO, 'TC',
                                                       countries='PHL',
                                                       year_range=(2013, 2013),
                                                       imp_str="Total Affected")
        # compare number of entries for all steps:
        self.assertEqual(len(df1.index), len(df2.index))
        self.assertEqual(impact_emdat.event_id.size, len(df1.index))
        # TC events in EM-DAT in the Philipppines, 2013:
        self.assertEqual(8, impact_emdat.event_id.size)
        # People affected by TC events in the Philippines in 2013 (AAI):
        self.assertAlmostEqual(17944571., impact_emdat.aai_agg, places=0)
        # People affected by Typhoon Hayian in the Philippines:
        self.assertAlmostEqual(1.610687e+07, impact_emdat.at_event[4], places=0)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestEmdatImport)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGDPScaling))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestEmdatProcessing))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestEmdatToImpact))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

init util
"""
import logging
from pint import UnitRegistry

from .config import *
from .constants import *
from .coordinates import *
from .save import *

ureg = UnitRegistry()

class log_level:
    """Context manager that sets all loggers with names starting with
    name_prefix (default is "") to a given specified level.

    Examples
    --------
    Set ALL loggers temporarily to the level 'WARNING'
    >>> with log_level(level='WARNING'):
    >>>     ...

    Set all the climada loggers temporarily to the level 'ERROR'
    >>> with log_level(level='ERROR', name_prefix='climada'):
    >>>     ...

    """

    def __init__(self, level, name_prefix=""):
        self.level = level
        self.loggers = {
            name: (logger, logger.level)
            for name, logger in logging.root.manager.loggerDict.items()
            if isinstance(logger, logging.Logger) and name.startswith(name_prefix)
            }
        if name_prefix == "":
            self.loggers[""] = (logging.getLogger(), logging.getLogger().level)

    def __enter__(self):
        for logger, _ in self.loggers.values():
            logger.setLevel(self.level)

    def __exit__(self, exception_type, exception, traceback):
        for logger, previous_level in self.loggers.values():
            logger.setLevel(previous_level)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test files_handler module.
"""

import unittest
from pathlib import Path

from climada.util.files_handler import to_list, get_file_names, download_file, \
get_extension
from climada.util.constants import DEMO_DIR, GLB_CENTROIDS_MAT, ENT_TEMPLATE_XLS


class TestDownloadUrl(unittest.TestCase):
    """Test download_file function"""
    def test_wrong_url_fail(self):
        """Error raised when wrong url."""
        url = 'https://ngdc.noaa.gov/eog/data/web_data/v4composites/F172012.v4.tar'
        try:
            with self.assertRaises(ValueError):
                download_file(url)
        except IOError:
            pass


class TestToStrList(unittest.TestCase):
    """Test to_list function"""
    def test_identity_pass(self):
        """Returns the same list if its length is correct."""
        num_exp = 3
        values = ['hi', 'ho', 'ha']
        val_name = 'values'
        out = to_list(num_exp, values, val_name)
        self.assertEqual(values, out)

    def test_one_to_list(self):
        """When input is a string or list with one element, it returns a list
        with the expected number of elments repeated"""
        num_exp = 3
        values = 'hi'
        val_name = 'values'
        out = to_list(num_exp, values, val_name)
        self.assertEqual(['hi', 'hi', 'hi'], out)

        values = ['ha']
        out = to_list(num_exp, values, val_name)
        self.assertEqual(['ha', 'ha', 'ha'], out)

    def test_list_wrong_length_fail(self):
        """When input is list of neither expected size nor one, fail."""
        num_exp = 3
        values = ['1', '2']
        val_name = 'values'

        with self.assertRaises(ValueError) as cm:
            to_list(num_exp, values, val_name)
        self.assertIn("Provide one or 3 values.", str(cm.exception))


class TestGetFileNames(unittest.TestCase):
    """Test get_file_names function. Only works with actually existing
        files and directories."""
    def test_one_file_copy(self):
        """If input is one file name, return a list with this file name"""
        file_name = GLB_CENTROIDS_MAT
        out = get_file_names(file_name)
        self.assertEqual([str(file_name)], out)

    def test_several_file_copy(self):
        """If input is a list with several file names, return the same list"""
        file_name = [GLB_CENTROIDS_MAT, ENT_TEMPLATE_XLS]
        out = get_file_names(file_name)
        self.assertEqual([str(x) for x in file_name], out)

    def test_folder_contents(self):
        """If input is one folder name, return a list with containg files.
        Folder names are not contained."""
        file_name = DEMO_DIR
        out = get_file_names(file_name)
        self.assertGreater(len(out), 0)
        for file in out:
            self.assertTrue(Path(file).is_file())

    def test_wrong_argument(self):
        """If the input contains a non-existing file, an empyt directory or a pattern that is not
        matched, the method should raise a ValueError."""
        empty_dir = DEMO_DIR.parent
        with self.assertRaises(ValueError) as ve:
            get_file_names(str(empty_dir))
        self.assertIn("no files", str(ve.exception))

        no_file = 'this is not a file'
        with self.assertRaises(ValueError) as ve:
            get_file_names(no_file)
        self.assertIn("cannot find", str(ve.exception))

    def test_globbing(self):
        """If input is a glob pattern, return a list of matching visible
            files; omit folders.
        """
        file_name = DEMO_DIR
        out = get_file_names(f'{file_name}/*')

        tmp_files = [str(f)
                     for f in Path(file_name).iterdir()
                     if f.is_file() and not f.name.startswith('.')]

        self.assertListEqual(sorted(tmp_files), sorted(out))


class TestExtension(unittest.TestCase):
    """Test get_extension"""

    def test_get_extension_no_pass(self):
        """Test no extension"""
        file_name = '/Users/aznarsig/Documents/Python/climada_python/data/demo/SC22000_VE__M1'
        self.assertEqual('', get_extension(file_name)[1])
        self.assertEqual(str(Path(file_name)), get_extension(file_name)[0])

    def test_get_extension_one_pass(self):
        """Test not compressed"""
        file_name = '/Users/aznarsig/Documents/Python/climada_python/data/demo/SC22000_VE__M1.grd'
        self.assertEqual('.grd', get_extension(file_name)[1])
        self.assertEqual(
            str(Path('/Users/aznarsig/Documents/Python/climada_python/data/demo/SC22000_VE__M1')),
            get_extension(file_name)[0])

    def test_get_extension_two_pass(self):
        """Test compressed"""
        file_name = '/Users/aznarsig/Documents/Python/climada_python' \
                    '/data/demo/SC22000_VE__M1.grd.gz'
        self.assertEqual('.grd.gz', get_extension(file_name)[1])
        self.assertEqual(
            str(Path('/Users/aznarsig/Documents/Python/climada_python/data/demo/SC22000_VE__M1')),
            get_extension(file_name)[0])


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestToStrList)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGetFileNames))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDownloadUrl))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestExtension))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.
Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.
CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.
---
Define functions to handle impact_yearsets
"""

import copy
import logging
import numpy as np
from numpy.random import default_rng

import climada.util.dates_times as u_dt

LOGGER = logging.getLogger(__name__)

def impact_yearset(imp, sampled_years, lam=None, correction_fac=True, seed=None):
    """Create a yearset of impacts (yimp) containing a probabilistic impact for each year
    in the sampled_years list by sampling events from the impact received as input with a
    Poisson distribution centered around lam per year (lam = sum(imp.frequency)).
    In contrast to the expected annual impact (eai) yimp contains impact values that
    differ among years. When correction factor is true, the yimp are scaled such
    that the average over all years is equal to the eai.

    Parameters
    -----------
        imp : climada.engine.Impact()
            impact object containing impacts per event
        sampled_years : list
            A list of years that shall be covered by the resulting yimp.
        seed : Any, optional
            seed for the default bit generator
            default: None

    Optional parameters
        lam: int
            The applied Poisson distribution is centered around lam events per year.
            If no lambda value is given, the default lam = sum(imp.frequency) is used.
        correction_fac : boolean
            If True a correction factor is applied to the resulting yimp. It is
            scaled in such a way that the expected annual impact (eai) of the yimp
            equals the eai of the input impact

    Returns
    -------
        yimp : climada.engine.Impact()
             yearset of impacts containing annual impacts for all sampled_years
        sampling_vect : 2D array
            The sampling vector specifies how to sample the yimp, it consists of one
            sub-array per sampled_year, which contains the event_ids of the events used to
            calculate the annual impacts.
            Can be used to re-create the exact same yimp.
      """

    n_sampled_years = len(sampled_years)

    #create sampling vector
    if not lam:
        lam = np.sum(imp.frequency)
    events_per_year = sample_from_poisson(n_sampled_years, lam, seed=seed)
    sampling_vect = sample_events(events_per_year, imp.frequency, seed=seed)

    #compute impact per sampled_year
    imp_per_year = compute_imp_per_year(imp, sampling_vect)

    #copy imp object as basis for the yimp object
    yimp = copy.deepcopy(imp)

    #save imp_per_year in yimp
    if correction_fac: #adjust for sampling error
        yimp.at_event = imp_per_year / calculate_correction_fac(imp_per_year, imp)
    else:
        yimp.at_event = imp_per_year

    #save calculations in yimp
    yimp.event_id = np.arange(1, n_sampled_years+1)
    yimp.tag['yimp object'] = True
    yimp.date = u_dt.str_to_date([str(date) + '-01-01' for date in sampled_years])
    yimp.frequency = np.ones(n_sampled_years)*sum(len(row) for row in sampling_vect
                                                            )/n_sampled_years

    return yimp, sampling_vect

def impact_yearset_from_sampling_vect(imp, sampled_years, sampling_vect, correction_fac=True):
    """Create a yearset of impacts (yimp) containing a probabilistic impact for each year
    in the sampled_years list by sampling events from the impact received as input following
    the sampling vector provided.
    In contrast to the expected annual impact (eai) yimp contains impact values that
    differ among years. When correction factor is true, the yimp are scaled such
    that the average over all years is equal to the eai.

    Parameters
    -----------
        imp : climada.engine.Impact()
            impact object containing impacts per event
        sampled_years : list
            A list of years that shall be covered by the resulting yimp.
        sampling_vect : 2D array
            The sampling vector specifies how to sample the yimp, it consists of one
            sub-array per sampled_year, which contains the event_ids of the events used to
            calculate the annual impacts.
            It needs to be obtained in a first call,
            i.e. [yimp, sampling_vect] = climada_yearsets.impact_yearset(...)
            and can then be provided in this function to obtain the exact same sampling
            (also for a different imp object)

    Optional parameter
        correction_fac : boolean
            If True a correction factor is applied to the resulting yimp. It is
            scaled in such a way that the expected annual impact (eai) of the yimp
            equals the eai of the input impact

    Returns
    -------
        yimp : climada.engine.Impact()
             yearset of impacts containing annual impacts for all sampled_years

    """

    #compute impact per sampled_year
    imp_per_year = compute_imp_per_year(imp, sampling_vect)

    #copy imp object as basis for the yimp object
    yimp = copy.deepcopy(imp)


    if correction_fac: #adjust for sampling error
        imp_per_year = imp_per_year / calculate_correction_fac(imp_per_year, imp)

    #save calculations in yimp
    yimp.at_event = imp_per_year
    n_sampled_years = len(sampled_years)
    yimp.event_id = np.arange(1, n_sampled_years+1)
    yimp.tag['yimp object'] = True
    yimp.date = u_dt.str_to_date([str(date) + '-01-01' for date in sampled_years])
    yimp.frequency = np.ones(n_sampled_years)*sum(len(row) for row in sampling_vect
                                                            )/n_sampled_years

    return yimp


def sample_from_poisson(n_sampled_years, lam, seed=None):
    """Sample the number of events for n_sampled_years

    Parameters
    -----------
        n_sampled_years : int
            The target number of years the impact yearset shall contain.
        lam: int
            the applied Poisson distribution is centered around lambda events per year
        seed : int, optional
            seed for numpy.random, will be set if not None
            default: None

    Returns
    -------
        events_per_year : array
            Number of events per sampled year
    """
    if seed is not None:
        np.random.seed(seed)
    if lam != 1:
        events_per_year = np.round(np.random.poisson(lam=lam,
                                                     size=n_sampled_years)).astype('int')
    else:
        events_per_year = np.ones(len(n_sampled_years))


    return events_per_year

def sample_events(events_per_year, freqs_orig, seed=None):
    """Sample events uniformely from an array (indices_orig) without replacement
    (if sum(events_per_year) > n_input_events the input events are repeated
    (tot_n_events/n_input_events) times, by ensuring that the same events doens't
    occur more than once per sampled year).

    Parameters
    -----------
        events_per_year : array
            Number of events per sampled year
        freqs_orig : array
            Frequency of each input event
        seed : Any, optional
            seed for the default bit generator.
            Default: None

    Returns
    -------
        sampling_vect : 2D array
            The sampling vector specifies how to sample the yimp, it consists of one
            sub-array per sampled_year, which contains the event_ids of the events used to
            calculate the annual impacts.
    """

    sampling_vect = []

    indices_orig = np.arange(len(freqs_orig))

    freqs = freqs_orig
    indices = indices_orig

    #sample events for each sampled year
    for amount_events in events_per_year:
        #if there are not enough input events, choice with no replace will fail
        if amount_events > len(freqs_orig):
            raise ValueError(f"cannot sample {amount_events} distinct events for a single year"
                             f" when there are only {len(freqs_orig)} input events")

        #add the original indices and frequencies to the pool if there are less events
        #in the pool than needed to fill the year one is sampling for
        #or if the pool is empty (not covered in case amount_events is 0)
        if len(np.unique(indices)) < amount_events or len(indices) == 0:
            indices = np.append(indices, indices_orig)
            freqs = np.append(freqs, freqs_orig)

        #ensure that each event only occurs once per sampled year
        unique_events = np.unique(indices, return_index=True)[0]
        probab_dis = freqs[np.unique(indices, return_index=True)[1]]/(
            np.sum(freqs[np.unique(indices, return_index=True)[1]]))

        #sample events
        rng = default_rng(seed)
        selected_events = rng.choice(unique_events, size=amount_events, replace=False,
                                     p=probab_dis).astype('int')

        #determine used events to remove them from sampling pool
        idx_to_remove = [np.where(indices == event)[0][0] for event in selected_events]
        indices = np.delete(indices, idx_to_remove)
        freqs = np.delete(freqs, idx_to_remove)

        #save sampled events in sampling vector
        sampling_vect.append(selected_events)

    return sampling_vect

def compute_imp_per_year(imp, sampling_vect):
    """Sample annual impacts from the given event_impacts according to the sampling dictionary

    Parameters
    -----------
        imp : climada.engine.Impact()
            impact object containing impacts per event
        sampling_vect : 2D array
            The sampling vector specifies how to sample the yimp, it consists of one
            sub-array per sampled_year, which contains the event_ids of the events used to
            calculate the annual impacts.

    Returns
    -------
        imp_per_year: array
            Sampled impact per year (length = sampled_years)
    """

    imp_per_year = [np.sum(imp.at_event[list(sampled_events)]) for sampled_events in
                    sampling_vect]

    return imp_per_year

def calculate_correction_fac(imp_per_year, imp):
    """Calculate a correction factor that can be used to scale the yimp in such
    a way that the expected annual impact (eai) of the yimp amounts to the eai
    of the input imp

    Parameters
    -----------
        imp_per_year : array
            sampled yimp
        imp : climada.engine.Impact()
            impact object containing impacts per event

    Returns
    -------
        correction_factor: int
            The correction factor is calculated as imp_eai/yimp_eai
    """

    yimp_eai = np.sum(imp_per_year)/len(imp_per_year)
    imp_eai = np.sum(imp.frequency*imp.at_event)
    correction_factor = imp_eai/yimp_eai
    LOGGER.info("The correction factor amounts to %s", (correction_factor-1)*100)

    # if correction_factor > 0.1:
    #     tex = raw_input("Do you want to exclude small events?")

    return correction_factor



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define MeasureSet class.
"""

__all__ = ['MeasureSet']

import copy
import logging
import ast
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xlsxwriter

from climada.entity.measures.base import Measure
from climada.entity.tag import Tag
import climada.util.hdf5_handler as u_hdf5

LOGGER = logging.getLogger(__name__)

DEF_VAR_MAT = {'sup_field_name': 'entity',
               'field_name': 'measures',
               'var_name': {'name': 'name',
                            'color': 'color',
                            'cost': 'cost',
                            'haz_int_a': 'hazard_intensity_impact_a',
                            'haz_int_b': 'hazard_intensity_impact_b',
                            'haz_frq': 'hazard_high_frequency_cutoff',
                            'haz_set': 'hazard_event_set',
                            'mdd_a': 'MDD_impact_a',
                            'mdd_b': 'MDD_impact_b',
                            'paa_a': 'PAA_impact_a',
                            'paa_b': 'PAA_impact_b',
                            'fun_map': 'damagefunctions_map',
                            'exp_set': 'assets_file',
                            'exp_reg': 'Region_ID',
                            'risk_att': 'risk_transfer_attachement',
                            'risk_cov': 'risk_transfer_cover',
                            'haz': 'peril_ID'
                           }
              }
"""MATLAB variable names"""

DEF_VAR_EXCEL = {'sheet_name': 'measures',
                 'col_name': {'name': 'name',
                              'color': 'color',
                              'cost': 'cost',
                              'haz_int_a': 'hazard intensity impact a',
                              'haz_int_b': 'hazard intensity impact b',
                              'haz_frq': 'hazard high frequency cutoff',
                              'haz_set': 'hazard event set',
                              'mdd_a': 'MDD impact a',
                              'mdd_b': 'MDD impact b',
                              'paa_a': 'PAA impact a',
                              'paa_b': 'PAA impact b',
                              'fun_map': 'damagefunctions map',
                              'exp_set': 'assets file',
                              'exp_reg': 'Region_ID',
                              'risk_att': 'risk transfer attachement',
                              'risk_cov': 'risk transfer cover',
                              'risk_fact': 'risk transfer cost factor',
                              'haz': 'peril_ID'
                             }
                }
"""Excel variable names"""

class MeasureSet():
    """Contains measures of type Measure. Loads from
    files with format defined in FILE_EXT.

    Attributes
    ----------
    tag : Tag
        information about the source data
    _data : dict
        cotains Measure classes. It's not suppossed to be
        directly accessed. Use the class methods instead.
    """

    def __init__(self):
        """Empty initialization.

        Examples
        --------
        Fill MeasureSet with values and check consistency data:

        >>> act_1 = Measure()
        >>> act_1.name = 'Seawall'
        >>> act_1.color_rgb = np.array([0.1529, 0.2510, 0.5451])
        >>> act_1.hazard_intensity = (1, 0)
        >>> act_1.mdd_impact = (1, 0)
        >>> act_1.paa_impact = (1, 0)
        >>> meas = MeasureSet()
        >>> meas.append(act_1)
        >>> meas.tag.description = "my dummy MeasureSet."
        >>> meas.check()

        Read measures from file and checks consistency data:

        >>> meas = MeasureSet.from_excel(ENT_TEMPLATE_XLS)
        """
        self.clear()

    def clear(self):
        """Reinitialize attributes."""
        self.tag = Tag()
        self._data = dict()  # {hazard_type : {name: Measure()}}

    def append(self, meas):
        """Append an Measure. Override if same name and haz_type.

        Parameters
        ----------
        meas : Measure
            Measure instance

        Raises
        ------
        ValueError
        """
        if not isinstance(meas, Measure):
            raise ValueError("Input value is not of type Measure.")
        if not meas.haz_type:
            LOGGER.warning("Input Measure's hazard type not set.")
        if not meas.name:
            LOGGER.warning("Input Measure's name not set.")
        if meas.haz_type not in self._data:
            self._data[meas.haz_type] = dict()
        self._data[meas.haz_type][meas.name] = meas

    def remove_measure(self, haz_type=None, name=None):
        """Remove impact function(s) with provided hazard type and/or id.
        If no input provided, all impact functions are removed.

        Parameters
        ----------
        haz_type : str, optional
            all impact functions with this hazard
        name : str, optional
            measure name
        """
        if (haz_type is not None) and (name is not None):
            try:
                del self._data[haz_type][name]
            except KeyError:
                LOGGER.info("No Measure with hazard %s and id %s.",
                            haz_type, name)
        elif haz_type is not None:
            try:
                del self._data[haz_type]
            except KeyError:
                LOGGER.info("No Measure with hazard %s.", haz_type)
        elif name is not None:
            haz_remove = self.get_hazard_types(name)
            if not haz_remove:
                LOGGER.info("No Measure with name %s.", name)
            for haz in haz_remove:
                del self._data[haz][name]
        else:
            self._data = dict()

    def get_measure(self, haz_type=None, name=None):
        """Get ImpactFunc(s) of input hazard type and/or id.
        If no input provided, all impact functions are returned.

        Parameters
        ----------
        haz_type : str, optional
            hazard type
        name : str, optional
            measure name

        Returns
        -------
        Measure (if haz_type and name),
        list(Measure) (if haz_type or name),
        {Measure.haz_type : {Measure.name : Measure}} (if None)
        """
        if (haz_type is not None) and (name is not None):
            try:
                return self._data[haz_type][name]
            except KeyError:
                LOGGER.info("No Measure with hazard %s and id %s.",
                            haz_type, name)
                return list()
        elif haz_type is not None:
            try:
                return list(self._data[haz_type].values())
            except KeyError:
                LOGGER.info("No Measure with hazard %s.", haz_type)
                return list()
        elif name is not None:
            haz_return = self.get_hazard_types(name)
            if not haz_return:
                LOGGER.info("No Measure with name %s.", name)
            meas_return = []
            for haz in haz_return:
                meas_return.append(self._data[haz][name])
            return meas_return
        else:
            return self._data

    def get_hazard_types(self, meas=None):
        """Get measures hazard types contained for the name provided.
        Return all hazard types if no input name.

        Parameters
        ----------
        name : str, optional
            measure name

        Returns
        -------
        list(str)
        """
        if meas is None:
            return list(self._data.keys())

        haz_return = []
        for haz, haz_dict in self._data.items():
            if meas in haz_dict:
                haz_return.append(haz)
        return haz_return

    def get_names(self, haz_type=None):
        """Get measures names contained for the hazard type provided.
        Return all names for each hazard type if no input hazard type.

        Parameters
        ----------
        haz_type : str, optional
            hazard type from which to obtain the names

        Returns
        -------
        list(Measure.name) (if haz_type provided),
        {Measure.haz_type : list(Measure.name)} (if no haz_type)
        """
        if haz_type is None:
            out_dict = dict()
            for haz, haz_dict in self._data.items():
                out_dict[haz] = list(haz_dict.keys())
            return out_dict

        try:
            return list(self._data[haz_type].keys())
        except KeyError:
            LOGGER.info("No Measure with hazard %s.", haz_type)
            return list()

    def size(self, haz_type=None, name=None):
        """Get number of measures contained with input hazard type and
        /or id. If no input provided, get total number of impact functions.

        Parameters
        ----------
        haz_type : str, optional
            hazard type
        name : str, optional
            measure name

        Returns
        -------
        int
        """
        if (haz_type is not None) and (name is not None) and \
        (isinstance(self.get_measure(haz_type, name), Measure)):
            return 1
        if (haz_type is not None) or (name is not None):
            return len(self.get_measure(haz_type, name))
        return sum(len(meas_list) for meas_list in self.get_names().values())

    def check(self):
        """Check instance attributes.

        Raises
        ------
        ValueError
        """
        for key_haz, meas_dict in self._data.items():
            def_color = plt.cm.get_cmap('Greys', len(meas_dict))
            for i_meas, (name, meas) in enumerate(meas_dict.items()):
                if (name != meas.name) | (name == ''):
                    raise ValueError("Wrong Measure.name: %s != %s."
                                     % (name, meas.name))
                if key_haz != meas.haz_type:
                    raise ValueError("Wrong Measure.haz_type: %s != %s."
                                     % (key_haz, meas.haz_type))
                # set default color if not set
                if np.array_equal(meas.color_rgb, np.zeros(3)):
                    meas.color_rgb = def_color(i_meas)
                meas.check()

    def extend(self, meas_set):
        """Extend measures of input MeasureSet to current
        MeasureSet. Overwrite Measure if same name and haz_type.

        Parameters
        ----------
        impact_funcs : MeasureSet
            ImpactFuncSet instance to extend

        Raises
        ------
        ValueError
        """
        meas_set.check()
        if self.size() == 0:
            self.__dict__ = copy.deepcopy(meas_set.__dict__)
            return

        self.tag.append(meas_set.tag)

        new_func = meas_set.get_measure()
        for _, meas_dict in new_func.items():
            for _, meas in meas_dict.items():
                self.append(meas)

    @classmethod
    def from_mat(cls, file_name, description='', var_names=DEF_VAR_MAT):
        """Read MATLAB file generated with previous MATLAB CLIMADA version.

        Parameters
        ----------
        file_name : str
            absolute file name
        description : str, optional
            description of the data
        var_names : dict, optional
            name of the variables in the file

        Returns
        -------
        meas_set: climada.entity.MeasureSet()
            Measure Set from matlab file
        """
        def read_att_mat(measures, data, file_name, var_names):
            """Read MATLAB measures attributes"""
            num_mes = len(data[var_names['var_name']['name']])
            for idx in range(0, num_mes):
                meas = Measure()

                meas.name = u_hdf5.get_str_from_ref(
                    file_name, data[var_names['var_name']['name']][idx][0])

                color_str = u_hdf5.get_str_from_ref(
                    file_name, data[var_names['var_name']['color']][idx][0])
                meas.color_rgb = np.fromstring(color_str, dtype=float, sep=' ')
                meas.cost = data[var_names['var_name']['cost']][idx][0]
                meas.haz_type = u_hdf5.get_str_from_ref(
                    file_name, data[var_names['var_name']['haz']][idx][0])
                meas.hazard_freq_cutoff = data[var_names['var_name']['haz_frq']][idx][0]
                meas.hazard_set = u_hdf5.get_str_from_ref(
                    file_name, data[var_names['var_name']['haz_set']][idx][0])
                try:
                    meas.hazard_inten_imp = (
                        data[var_names['var_name']['haz_int_a']][idx][0],
                        data[var_names['var_name']['haz_int_b']][0][idx])
                except KeyError:
                    meas.hazard_inten_imp = (
                        data[var_names['var_name']['haz_int_a'][:-2]][idx][0], 0)

                # different convention of signes followed in MATLAB!
                meas.mdd_impact = (data[var_names['var_name']['mdd_a']][idx][0],
                                   data[var_names['var_name']['mdd_b']][idx][0])
                meas.paa_impact = (data[var_names['var_name']['paa_a']][idx][0],
                                   data[var_names['var_name']['paa_b']][idx][0])
                meas.imp_fun_map = u_hdf5.get_str_from_ref(
                    file_name, data[var_names['var_name']['fun_map']][idx][0])

                meas.exposures_set = u_hdf5.get_str_from_ref(
                    file_name, data[var_names['var_name']['exp_set']][idx][0])
                exp_region_id = data[var_names['var_name']['exp_reg']][idx][0]
                if exp_region_id:
                    meas.exp_region_id = [exp_region_id]
                meas.risk_transf_attach = data[var_names['var_name']['risk_att']][idx][0]
                meas.risk_transf_cover = data[var_names['var_name']['risk_cov']][idx][0]

                measures.append(meas)

        data = u_hdf5.read(file_name)
        meas_set = cls()
        meas_set.tag.file_name = str(file_name)
        meas_set.tag.description = description
        try:
            data = data[var_names['sup_field_name']]
        except KeyError:
            pass

        try:
            data = data[var_names['field_name']]
            read_att_mat(meas_set, data, file_name, var_names)
        except KeyError as var_err:
            raise KeyError("Variable not in MAT file: " + str(var_err)) from var_err

        return meas_set

    def read_mat(self, *args, **kwargs):
       """This function is deprecated, use MeasureSet.from_mat instead."""
       LOGGER.warning("The use of MeasureSet.read_mat is deprecated."
                      "Use MeasureSet.from_mat instead.")
       self.__dict__ = MeasureSet.from_mat(*args, **kwargs).__dict__

    @classmethod
    def from_excel(cls, file_name, description='', var_names=DEF_VAR_EXCEL):
        """Read excel file following template and store variables.

        Parameters
        ----------
        file_name : str
            absolute file name
        description : str, optional
            description of the data
        var_names : dict, optional
            name of the variables in the file

        Returns
        -------
        meas_set : climada.entity.MeasureSet
            Measures set from Excel
        """
        def read_att_excel(measures, dfr, var_names):
            """Read Excel measures attributes"""
            num_mes = len(dfr.index)
            for idx in range(0, num_mes):
                meas = Measure()

                meas.name = dfr[var_names['col_name']['name']][idx]
                try:
                    meas.haz_type = dfr[var_names['col_name']['haz']][idx]
                except KeyError:
                    pass
                meas.color_rgb = np.fromstring(
                    dfr[var_names['col_name']['color']][idx], dtype=float, sep=' ')
                meas.cost = dfr[var_names['col_name']['cost']][idx]

                meas.hazard_freq_cutoff = dfr[var_names['col_name']['haz_frq']][idx]
                meas.hazard_set = dfr[var_names['col_name']['haz_set']][idx]
                # Search for (a, b) values, put a = 1 otherwise
                try:
                    meas.hazard_inten_imp = (dfr[var_names['col_name']['haz_int_a']][idx],
                                             dfr[var_names['col_name']['haz_int_b']][idx])
                except KeyError:
                    meas.hazard_inten_imp = (1, dfr['hazard intensity impact'][idx])

                try:
                    meas.exposures_set = dfr[var_names['col_name']['exp_set']][idx]
                    meas.exp_region_id = ast.literal_eval(dfr[var_names['col_name']['exp_reg']][idx])
                except KeyError:
                    pass
                except ValueError:
                    meas.exp_region_id = dfr[var_names['col_name']['exp_reg']][idx]

                meas.mdd_impact = (dfr[var_names['col_name']['mdd_a']][idx],
                                   dfr[var_names['col_name']['mdd_b']][idx])
                meas.paa_impact = (dfr[var_names['col_name']['paa_a']][idx],
                                   dfr[var_names['col_name']['paa_b']][idx])
                meas.imp_fun_map = dfr[var_names['col_name']['fun_map']][idx]
                meas.risk_transf_attach = dfr[var_names['col_name']['risk_att']][idx]
                meas.risk_transf_cover = dfr[var_names['col_name']['risk_cov']][idx]
                try:
                    meas.risk_transf_cost_factor = dfr[var_names['col_name']['risk_fact']][idx]
                except KeyError:
                    pass

                measures.append(meas)

        dfr = pd.read_excel(file_name, var_names['sheet_name'])
        dfr = dfr.fillna('')
        meas_set = cls()
        meas_set.tag.file_name = str(file_name)
        meas_set.tag.description = description
        try:
            read_att_excel(meas_set, dfr, var_names)
        except KeyError as var_err:
            raise KeyError("Variable not in Excel file: " + str(var_err)) from var_err

        return meas_set

    def read_excel(self, *args, **kwargs):
        """This function is deprecated, use MeasureSet.from_excel instead."""
        LOGGER.warning("The use ofMeasureSet.read_excel is deprecated."
                       "Use MeasureSet.from_excel instead.")
        self.__dict__ = MeasureSet.from_excel(*args, **kwargs).__dict__

    def write_excel(self, file_name, var_names=DEF_VAR_EXCEL):
        """Write excel file following template.

        Parameters
        ----------
        file_name : str
            absolute file name to write
        var_names : dict, optional
            name of the variables in the file
        """
        def write_meas(row_ini, imp_ws, xls_data):
            """Write one measure"""
            for icol, col_dat in enumerate(xls_data):
                imp_ws.write(row_ini, icol, col_dat)

        meas_wb = xlsxwriter.Workbook(file_name)
        mead_ws = meas_wb.add_worksheet(var_names['sheet_name'])

        header = [var_names['col_name']['name'], var_names['col_name']['color'],
                  var_names['col_name']['cost'], var_names['col_name']['haz_int_a'],
                  var_names['col_name']['haz_int_b'], var_names['col_name']['haz_frq'],
                  var_names['col_name']['haz_set'], var_names['col_name']['mdd_a'],
                  var_names['col_name']['mdd_b'], var_names['col_name']['paa_a'],
                  var_names['col_name']['paa_b'], var_names['col_name']['fun_map'],
                  var_names['col_name']['exp_set'], var_names['col_name']['exp_reg'],
                  var_names['col_name']['risk_att'], var_names['col_name']['risk_cov'],
                  var_names['col_name']['haz']]
        for icol, head_dat in enumerate(header):
            mead_ws.write(0, icol, head_dat)
        for row_ini, (_, haz_dict) in enumerate(self._data.items(), 1):
            for meas_name, meas in haz_dict.items():
                xls_data = [meas_name, ' '.join(list(map(str, meas.color_rgb))),
                            meas.cost, meas.hazard_inten_imp[0],
                            meas.hazard_inten_imp[1], meas.hazard_freq_cutoff,
                            meas.hazard_set, meas.mdd_impact[0], meas.mdd_impact[1],
                            meas.paa_impact[0], meas.paa_impact[1], meas.imp_fun_map,
                            meas.exposures_set, str(meas.exp_region_id), meas.risk_transf_attach,
                            meas.risk_transf_cover, meas.haz_type]
            write_meas(row_ini, mead_ws, xls_data)
        meas_wb.close()
